module wei_obj_D
  use wei_util 
  use wei_down
  use wei_run
  use wei_external
  use wei_type
  !---------------
  implicit none
contains
  !--------------------------------------------------------------
  !
  !  Function to initialize a data object either from
  ! 
  !    a file or from then general parameters
  !
  !
  !
  logical function wei_data_init(X,tag)
    type(wedata),       intent(inout) :: X
    character(len=128), intent(in)    :: tag

    wei_data_init=F

    X%tag = tag
    if(isfile(X%tag)) then
       if(.not. data_history_init(X)) then
         write(0,*) "trouble initializing data_history"
         return
       end if
    else
       if(.not. data_param_init(X)) then
         write(0,*) "trouble initializing data_parameter"
         return
       end if
    end if

    wei_data_init=T

  end function wei_data_init
  !--------------------------------------------------------------
  !
  !  Function to initialize source object from input or 
  !          construct a wavelet
  ! 
  logical function wei_source_init(X,tag)
    type(wedata),       intent(inout) :: X
    character(len=128), intent(in)    :: tag

    wei_source_init=F

    X%tag = tag
    if(isfile(X%tag)) then
       if(.not. source_history_init(X) ) then
         write(0,*) "trouble initializing data_history"
         return
       end if
    else
!       if(.not. source_param_init(X)) then
!         write(0,*) "trouble making source function"
!         return
!       end if
       call erexit("Source module not built")
    end if

    wei_source_init=T

  end function wei_source_init

  !----------------------------------------------------------------
  !
  !
  !  Create an auxilary data file
  !
  logical function wei_aux_data_init(X,tag)
    type(wedata),       intent(inout) :: X
    character(len=128), intent(in)    :: tag

    wei_aux_data_init=F

    X%tag = tag
    if(.not. data_scratch_init(X)) then
       write(0,*) "trouble initializing scratch data file"
       return
    end if

    wei_aux_data_init=T

  end function wei_aux_data_init
  !----------------------------------------------------------------
  !
  !  Initialize a tag from history
  !   check to make sure that is dimensions don't disagree
  !    with the general params
  ! 
  logical function data_history_init(X)
    type(wedata), intent(inout) :: X
    integer:: bogus
#ifdef DEBUG
    character(len=128), parameter :: unit='data_history_init'
    if(run%debug) call in(unit)
#endif

    call auxinout(X%tag)
    data_history_init=F

    X%from = "history"
    call from_aux(X%tag,'esize',X%esize,4)
    if(X%esize==4) call seperr("data has wrong esize, input frequency")
!! . . Initialize Axes calls from history file
    call hinaxis(X%tag,X%amx,1); X%amx%l='mx'
    call hinaxis(X%tag,X%amy,2); X%amy%l='my'
    call hinaxis(X%tag,X%aw ,3); X%aw%l='w_'
    call hinaxis(X%tag,X%asx,4); X%asx%l='sx'
    call hinaxis(X%tag,X%asy,5); X%asy%l ='sy'

!! . . geometry check
    if((X%amx%d .lt. 0) .or. (X%amy%d .lt. 0))&
    call seperr("Please refrain from negative d1 and/or d2")
!! . . Default indices to read into prop. grid
    X%indx=1;                      X%indy=1
    X%finx=X%indx + X%amx%n -1;    X%finy=X%indy + X%amy%n -1

    call down_global_var_calc(X)

    if(.not. data_setup(X,.false.)) then
       write(0,*) "trouble setting up data"
       return
    end if

    if(run%verb) call data_report(X)
    data_history_init=T
#ifdef DEBUG
    if(run%debug) call data_report(X)
    if(run%debug) call out(unit)
#endif
  end function data_history_init
  !----------------------------------------------------------------
  !
  ! almost the same as data_history_init, leaving out assigns to 
  !     down structure
  logical function source_history_init(X)
    type(wedata), intent(inout) :: X
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='data_source_init'
    if(run%debug) call in(unit)
#endif
    source_history_init=F
    call auxinout(X%tag)
    X%from = "history"
    call from_aux(X%tag,'esize',X%esize,4)
    if(X%esize==4) call seperr("source has wrong esize, input frequency")
!! . . Initialize Axes calls from history file
if(run%verb)write(0,*) 'fatx', down%fatx
    X%amx = down%fatx!    call hinaxis(X%tag,X%amx,1)
    X%amy = down%faty!    call hinaxis(X%tag,X%amy,2)
    call hinaxis(X%tag,X%aw ,3)

    X%asx%n=1;X%asx%d=1.;X%asx%o=0. ! need default values
    X%asy%n=1;X%asy%d=1.;X%asy%o=0. ! need default values
!! Frequency comparisson to data file
    if (X%aw%n /= down%aw%n) call erexit("Source file wrong length")   
    if (X%aw%d /= down%aw%d) call erexit("Source file freq. increment wrong")
    if (X%aw%o /= down%aw%o) call erexit("Source file freq. origin wrong")

!! . . Default indices to read into prop. grid
    X%indx=1;                      X%indy=1
    X%finx=X%indx + X%amx%n -1;    X%finy=X%indy + X%amy%n -1

    if(.not. data_setup(X,.false.)) then
       write(0,*) "trouble setting up source files"
       return
    end if

    !!re-allocate array reallocte array after distribution to size of
    !!propagation grid
    deallocate(X%bin)
    allocate(X%bin(down%amx%n,down%amy%n,down%nws),stat=er)
    if(er/=0) call seperr('trouble at source_history_init reallocate D_d.bin')
    X%bin=(0.0,0.0)

    if (down%antion) then
       allocate(X%imgbin(down%amx%n,down%amy%n,down%nws),stat=er)    
       if(er/=0) call seperr('trouble at source_history_init reallocate D_d.imgbin')
       X%imgbin=X%bin !(0.0,0.0)
    else
       X%imgbin=>X%bin
    end if
    if(run%verb) call data_report(X)
    source_history_init=T
#ifdef DEBUG
!    if(run%debug) call data_report(X)
    if(run%debug) call out(unit)
#endif
  end function source_history_init
!----------------------------------------------------------------
!      
!   relocate wavefield within the model
!    remember: amx%o is a relative distance from the shot!
  subroutine data_position(X,isx,isy)
    type(wedata),intent(in) :: X
    integer :: isx,isy

       down%amx%o= min( X%amx%o + X%asx%o+X%asx%d*(isx-1), X%asx%o+X%asx%d*(isx-1) ) 
       down%amy%o= min( X%amy%o + X%asy%o+X%asy%d*(isy-1), X%asy%o+X%asy%d*(isy-1) ) 
     end subroutine data_position
!----------------------------------------------------------------
  subroutine data_position_report()
    if (run%verb)write(0,'(a,f9.3)')"         *  Working grid positioned at amx%o=",&
                                     down%amx%o
    if (run%verb)write(0,'(a,f9.3)')"         *                             amy%o=",&
                                     down%amy%o
  end subroutine data_position_report
!----------------------------------------------------------------
  !
  !
  !
  logical function create_data_file(X,tag)
    type(wedata), intent(in) :: X
    character(len=*)         :: tag
       create_data_file=F 
       call auxscr(tag)
       call to_history("esize",8,tag)
       call houaxes(tag,X%amx,X%amy,X%aw,X%asx,X%asy)
       create_data_file=T 
   end function 
  !----------------------------------------------------------------
  logical function data_param_init(X)
    type(wedata), intent(inout) :: X
#ifdef DEBUG
    character(len=128), parameter :: unit='data_param_init' 
    if(run%debug) call in(unit)
#endif
    data_param_init=F

    X%from = "parameter"
    X%amx = down%amx; X%amy = down%amy; 
    X%asx = down%asx; X%asy = down%asy; 
    X%aw  = down%aw;  
    if(.not. create_data_file(X,X%tag)) then
      write(0,*) "trouble creating data file"
      return
    end if
    if(.not. data_setup(X,.true.)) then
      write(0,*) "trouble setting up data"
      return
    end if

    data_param_init=T
#ifdef DEBUG
    if(run%debug) call data_report(X)
    if(run%debug) call out(unit)
#endif
  end function data_param_init

!----------------------------------------------------------------
  subroutine data_assign(X,val)
   type(wedata) :: X
   complex :: val
   logical ::a

    a=file_val_assign(X%tag,val)
  end subroutine data_assign
  !----------------------------------------------------------------
  !
  ! data_scratch_init
  !
  !  Create a data scratch file
  !
  !
  logical function data_scratch_init(X)
    type(wedata), intent(inout) :: X
#ifdef DEBUG
    character(len=128), parameter :: unit='data_scratch_init' 
    if(run%debug) call in(unit)
#endif

    data_scratch_init=F
    X%from = "parameter"
    X%amx = down%amx; X%amy = down%amy; 
    X%asx = down%asx; X%asy = down%asy
    X%aw  = down%aw; 

    X%indx=1;                      X%indy=1
    X%finx=X%indx + X%amx%n -1;    X%finy=X%indy + X%amy%n -1

    if(.not. create_data_file(X,X%tag)) then
      write(0,*) "trouble  creating data file from scratch"
      return
    end if
    if(.not. data_setup(X,.true.)) then
      write(0,*) "trouble setting up data from scratch"
      return
    end if
    data_scratch_init=T
    if(run%verb)call data_report(X)
#ifdef DEBUG
    if(run%debug) call data_report(X)
    if(run%debug) call out(unit)
#endif
  end function data_scratch_init
!----------------------------------------------------------------  
  subroutine data_passive_pointer(X,Y)
    type (wedata),intent(out):: X 
    type (wedata),intent(in) :: Y


    down%D_d = down%D; 
    X%bites= Y%bites
    X%tag  = Y%tag; X%tag_0   = Y%tag_0
    X%amx  = Y%amx; X%amy     = Y%amy  
    X%asx  = Y%asx; X%asy     = Y%asy
    X%aw   = Y%aw ; X%aw_local= Y%aw_local
    X%bin  =>Y%bin
  end subroutine data_passive_pointer
  !----------------------------------------------------------------
  ! 
  !  The core  routine to setup data
  logical function data_setup(X,zero)
    type(wedata), intent(inout) :: X
    logical :: zero
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='data_setup' 
    if(run%debug) call in(unit)
#endif
    data_setup=F
    X%bites = 8* product( (/X%amx%n,X%amy%n,down%nws/) )

    !! FREQUENCY blocking
    if( (1.0*X%aw_local%n)/down%nws /= X%aw_local%n/down%nws ) then
       call data_report(X)
       call seperr('D.aw.n is not a multiple of nws')
    end if
    down%nwb = down%nw_local 

    !allocate array
!      allocate(X%bin(down%amx%n,down%amy%n,down%nws),stat=er)
      allocate(X%bin(X%amx%n,X%amy%n,down%nws),stat=er)
      if(er/=0) call seperr('cant allocate D.bin')
      X%bin=(0.0,0.0)
      if (down%antion) then
         allocate(X%imgbin(X%amx%n,X%amy%n,down%nws),stat=er)    
         if(er/=0) call seperr('cant allocate D.imgbin')
         X%imgbin=X%bin
      else
         X%imgbin=>X%bin
      end if


    if(zero .and. run%m_nth==1) then
      if(.not. file_val_assign(X%tag,cmplx(0.,0.))) then
        write(0,*) "trouble setting ",trim(X%tag_0)
        return
      end if
    end if
   if(run%mpi) then
     if(.not.  mpi_data_setup(X,zero)) then
       write(0,*) "trouble initializng data mpi"
       return
     end if
   else
     X%aw_local=X%aw
   end if

   data_setup=T
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function data_setup

  !----------------------------------------------------------------
  !
  !   Setup data structure for local files if
  !     we are running MPI

  logical function mpi_data_setup(X,zero)
    type(wedata), intent(inout) :: X
    integer ::n
    logical :: zero
    mpi_data_setup=F

    X%tag_0=X%tag
    write(X%tag,"(a,a)") trim(X%tag_0),run%tag_num

    X%aw_local%n=down%nw_local*down%nws

    n=X%aw%n
    X%aw%n=down%nw_local*down%nws
    
    if(.not. create_data_file(X,X%tag)) then
       write(0,*) "trouble setting up mpi data"
       return
    end if
    X%aw%n=n
       
    if(.not. zero) then 
       call distribute_data(X)
    else
       !       if(.not. file_val_assign(X%tag,cmplx(0.,0.))) then
       !          write(0,*) "trouble setting ",trim(X%tag_0)
       !          return
       !       end if
    end if
       
    mpi_data_setup=T
  end function mpi_data_setup

  !----------------------------------------------------------------
  !
  ! Information on the data
  !
  !
  subroutine data_report(X)
    type(wedata), intent(in) :: X
    write(0,*)'        ************************************************'
    write(0,*)'        *       ',trim(X%tag),' from ',trim(X%from)
    write(0,'(a,i3,f9.3,f9.3)       ') "         *   Offx: ",X%amx%n, X%amx%o, X%amx%d
    write(0,'(a,i3,f9.3,f9.3)       ') "         *   Offy: ",X%amy%n, X%amy%o, X%amy%d
    write(0,'(a,i3,f9.3,f9.3)       ') "         *   Shtx: ",X%asx%n, X%asx%o, X%asx%d
    write(0,'(a,i3,f9.3,f9.3)       ') "         *   Shty: ",X%asy%n, X%asy%o, X%asy%d
    write(0,'(a,i3,a,i3,a,f5.3,f9.3,a,i3,a)') "         *     W: ",X%aw%n,&
    "(",X%aw_local%n,")",  X%aw%o,  X%aw%d," ->",down%nwb," blocks"
    if(associated(X%bin)) then
       write(0,'(a,f9.2,a,f9.2,a)        ') "         * block: ",X%bites/1024./1024.," / ",8*size(X%bin)/1024./1024.," Mbytes"
    end if
  end subroutine data_report
  !--------------------------------------------------------------
  ! 
  ! Read the data 
  ! 
  !
  logical function data_read(X,k)
    type(wedata), intent(inout) :: X
    integer, intent(in) :: k
    integer :: stat
#ifdef DEBUG
    character(len=128), parameter :: unit='data_read'
    if(run%debug) call in(unit)
#endif
    data_read=F
    X%bin=(0.0,0.0)

    stat=sseek_block(X%tag,k,X%bites,0)
    if(X%bites/= sreed(X%tag, X%bin(X%indx:X%finx,X%indy:X%finy,:), X%bites)) then
       write(0,*) "Badness for file ",trim(X%tag)
       write(0,*) "trouble reading block",k,'machine',run%m_ith  
    else
!       X%bin=X%bin/1000
      data_read=T
    end if

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function data_read
  !--------------------------------------------------------------
  ! figure out where the source goes in the work space
  subroutine data_indices_land(X,isx,isy)
    type(wedata) ::X
    integer::isx,isy
#ifdef DEBUG
    character(len=128), parameter :: unit='calc_indices'
    if(run%debug) call in(unit)
#endif 
    X%indx= 1; X%indy= 1;

if(.not.down%passive) then
   if(X%amx%n .lt. down%amx%n/2 +1 )then
      X%indx= (1.5+ ((down%asx%o + (isx-1)*down%asx%d) - down%amx%o)/down%amx%d) -(X%amx%n-1)/2
   end if
   if(X%amy%n .lt. down%amy%n/2 +1 )then
      X%indy= (1.5+ ((down%asy%o + (isy-1)*down%asy%d) - down%amy%o)/down%amy%d) -(X%amy%n-1)/2
   end if
end if
    X%finx=X%indx + X%amx%n -1
    X%finy=X%indy + X%amy%n -1

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine data_indices_land
!----------------------------------------------------------------
! figure out where the source goes in the work space
  subroutine data_indices_marine(A,B)
    type(wedata) ::A,B
  !! remember: am_%o is a relative distance from the shot!
#ifdef DEBUG
    character(len=128), parameter :: unit='calc_indices_marine'
    if(run%debug) call in(unit)
#endif
    A%indx=1;    B%indx=1;    A%indy=1;    B%indy=1
    if( .not.down%passive) then
       if (A%amx%n .gt. 1 .and. B%amx%n == 1) then
          if (A%amx%o .gt. 0)then 
             A%indx= 1.5+    (A%amx%o)/down%amx%d 
          else
             B%indx= 1.5+ abs(A%amx%o)/down%amx%d
          end if
       else
          B%indx=1
       end if

       if (A%amy%n .gt. 1 .and. B%amy%n == 1) then
          if (A%amy%o .gt. 0)then
             A%indy= 1.5+    (A%amy%o)/down%amy%d 
          else
             B%indy= 1.5+ abs(A%amy%o)/down%amy%d 
          end if
       else
          B%indy=1
       end if
    end if

    A%finx=A%indx + A%amx%n -1
    A%finy=A%indy + A%amy%n -1
    B%finx=B%indx + B%amx%n -1
    B%finy=B%indy + B%amy%n -1

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine data_indices_marine
!----------------------------------------------------------------

!----------------------------------------------------------------
subroutine source_report(X)
type(wedata)::X
if (X%amx%n==1) then
   if(run%verb)write(0,'(a,f9.3)') '         *  Source located at x=',&
            down%amx%o + (X%indx-1)*down%amx%d +((X%amx%n-1)/2)*down%amx%d
else
   if(run%verb)write(0,'(a,f9.3)') '         *  Source located at x=',&
            down%amx%o + X%amx%o
end if
if  (X%amy%n==1) then
if(run%verb)write(0,'(a,f9.3)') '         *                    y=',&
            down%amy%o + (X%indy-1)*down%amy%d +((X%amy%n-1)/2)*down%amy%d
else
   if(run%verb)write(0,'(a,f9.3)') '         *                 y=',&
            down%amy%o + X%amy%o
end if


end subroutine source_report
 !----------------------------------------------------------------
  logical function data_write(X,k)
    type(wedata), intent(inout) :: X
    integer, intent(in) :: k
    integer :: stat
#ifdef DEBUG
    character(len=128), parameter :: unit='data_write' 
    if(run%debug) call in(unit)
#endif
    data_write=F 

    stat=sseek_block(X%tag,k-1,X%bites,0)
    if(X%bites/= srite(X%tag, X%bin, X%bites)) then
      write(0,*) "trouble writing data block",k
      return
    end if

     call to_history("ifreq",k,X%tag)
 
    data_write=T 
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function data_write
  !----------------------------------------------------------------
  !
  ! 
  !
  !
  subroutine distribute_data(X)
    type(wedata), intent(inout) :: X
    integer, external :: sseek
    integer :: ierr,i,j
#ifdef SEP_MPI
    if(run%mpi) then
       if(run%m_ith==0) ierr= sseek(trim(X%tag_0),0,0)
       ierr=sseek(trim(X%tag),0,0)
       do j=1, X%asy%n
          do i=1, X%asx%n
             if(0/=mpi_sep_tag_distribute(0,X%tag_0,run%m_ith,X%tag,X%bites/4,&
             down%nwp, down%iw_thread)) &
             call seperr("trouble distributing data ");
          end do
       end do
    end if
    ierr=sseek(trim(X%tag),0,0)
#endif
  end subroutine distribute_data
  !----------------------------------------------------------------
  subroutine combine_data(X)
    type(wedata), intent(inout) :: X
    integer :: ierr,i
    integer, external :: sseek
#ifdef SEP_MPI
    i=0;if(down%add) i=1
    if( .not. down%adj .and. run%mpi) then
       if(run%m_ith==0) ierr= sseek(trim(X%tag_0),0,0)
       ierr=sseek(trim(X%tag),0,0)

       if(0/=mpi_sep_tag_combine(run%m_ith,X%tag,0,X%tag_0, &
       X%bites/4,down%nwp,down%iw_thread,i)) &
       call seperr("trouble combining data ");
    end if
    call weiline()
!    write(0,*) "        *         UNLINKING ",trim(X%tag)
    call aux_unlink(trim(X%tag))
#endif
  end subroutine combine_data

  !----------------------------------------------------------------
  !
  !  Copy data
  !
  !
  subroutine data_copy(X,Y)
    type(wedata), intent(inout) :: X
    type(wedata), intent(inout) :: Y
    integer :: iwb
    logical :: res
#ifdef DEBUG
    character(len=128), parameter :: unit='data_copy'
    if(run%debug) call in(unit)
#endif

    do iwb=1,down%nwb
       res= data_read (X,iwb-1)
       Y%bin = X%bin
       res= data_write(Y,iwb)
    end do

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
   end subroutine data_copy

  !----------------------------------------------------------------
  ! 
  !  Add two data structures
  ! 
  ! 
  subroutine data_add(X,Y)
    type(wedata), intent(inout) :: X
    type(wedata), intent(inout) :: Y
    integer :: iwb
    logical :: res
#ifdef DEBUG
    character(len=128), parameter :: unit='data_add'
    if(run%debug) call in(unit)
#endif

    do iwb=1,down%nwb
       res= data_read (X,iwb)
       res= data_read (Y,iwb)
       Y%bin = Y%bin + X%bin
       res= data_write(Y,iwb)
    end do

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine data_add

  !----------------------------------------------------------------  

end module wei_obj_D





