module wei_obj_R
  use wei_util
  use wei_external
  use wei_run
  use wei_down
  use wei_down_type
  use wei_type
  use wei_lagrange

  implicit none
  integer, private, save,pointer  :: image_master(:)
  integer, private, save  :: image_comm_group
  logical, private, save  :: image_lead_thread


contains
  !----------------------------------------------------------------
  !
  !
  ! Image initializing routine
  !
  !
  logical function  wei_image_init(X,tag)
    type(weimag),       intent(inout) :: X
    character(len=128), intent(in)    :: tag
    integer ::ierr

    wei_image_init=F
    X%tag = tag

!    call MPI_Barrier(MPI_COMM_WORLD,ierr)
    if(((.not.down%adj) .or. (down%add)) .and. (isfile(X%tag))) then
       if(.not. image_history_init(X)) then
         write(0,*) "trouble initializing image_history"
         return
       end if       
    else
       if(.not. image_param_init(X)) then  !THIS ONE
         write(0,*) "trouble initializing image_param"
         return
       end if       
    end if

    if(var%use_disk_buf) then
       write(X%tag_var,"(a,a)") "R_var",run%tag_num
       call auxinout(X%tag_var)
       ierr= srite(X%tag_var,0.,4)
    end if

    wei_image_init=T
     
  end function wei_image_init

  !----------------------------------------------------------------
  !
  ! Initialize an image file based on parameters
  !
  !
  logical function image_param_init(X)
    type(weimag), intent(inout) :: X
    logical :: image_real
    integer :: noffx,noffy
#ifdef DEBUG
    character(len=128), parameter :: unit='image_param_init'
    if(run%debug) call in(unit)
#endif
    image_param_init=F
    X%from = "parameter"
!! . . depth o-rama
    X%az%d=down%az%d
    X%az%n=down%az%n-down%firstz+1
    X%az%o=down%az%o + down%firstz*down%az%d

!! . . calculate origin for CIG's
    down%ahx%o= -( down%ahx%n -1)/2. * down%amx%d
    down%ahy%o= -( down%ahy%n -1)/2. * down%amy%d !offset origin begins w/ neg. offset
    down%ahx%d=down%amx%d
    down%ahy%d=down%amy%d
    down%ahx%l='hx'
    down%ahy%l='hy'

    X%ahx=down%ahx
    X%ahy=down%ahy
!!total image size from earth covered by recievers
    X%aix=down%aix 
    X%aiy=down%aiy
!!propagation image size=size of 1 shotgather

    X%amx=down%amx 
    X%amy=down%amy
!!frequecy axis?
    if(down%image_freqs) X%aw=down%aw; X%az%l='z_'
    
    call from_param("image_real",image_real,.true.)
    X%esize=4;if(.not.image_real) X%esize=8

     
    if(run%m_ith==0) then 
      if(.not. create_image_file(X,X%tag)) then
        write(0,*) "trouble creating image"
        return
       end if
    end if

    if(.not. image_setup(X,.true.)) then
      write(0,*) "trouble in image setup"
      return
    end if


     image_param_init=T
    if(run%debug .or. run%verb) call image_report(X)
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function image_param_init
  !----------------------------------------------------------------
  !
  ! Write out parameters for an image file
  !
  !
  logical function create_image_file(X,tag)
    type(weimag) :: X
    type(wedata)::D,D_d
    type(weslow)::S
    type(wewfld)::W,W_d
    logical :: i_real
    integer :: i
    character(len=*) ::  tag 
    create_image_file=F

    call auxscr(tag)
    call to_history("esize",X%esize,tag)
    if (down%image_freqs)then
       call houaxes(tag,X%aix,X%aiy,X%ahx,X%ahy,X%aw,X%az)
    else
       call houaxes(tag,X%aix,X%aiy,X%ahx,X%ahy,X%az)
    end if
    !! General info into
    !call to_history("D" ,trim(D%tag_0) ,tag)
    !call to_history("D_d",trim(D_d%tag_0) ,tag)
    !call to_history("S" ,trim(S%s_s%tag_0) ,tag)
    !if (down%wIOu) call to_history("W" ,trim(W%tag) ,tag)
    !if (down%wIOd) call to_history("W_d" ,trim(W_d%tag) ,tag)
    call to_history("nref",down%nref ,tag)
    call to_history("passive",down%passive ,tag)
    call to_history("velocity",S%s_s%velocity ,tag)
    call to_history("operation",run%operation ,tag)
    call to_history("adj",down%adj ,tag)
    call to_history("add", down%add,tag)
    call to_history("order",down%order ,tag)
    call to_history("conv_wave",down%conv_wave ,tag)
    call to_history("image_freqs",down%image_freqs,tag)
    call to_history("firstz",down%firstz,tag)
    call to_history("antlialiasing",down%antion,tag)
    if (down%antion) then
       call to_history("kmax_x",down%pmax_x,tag)
       call to_history("kmax_y",down%pmax_y,tag)
       call to_history("kmin_x",down%pmin_x,tag)
       call to_history("kmin_y",down%pmin_y,tag)
    end if
    call to_history("multiple subtraction",down%multsub,tag)
    call to_history("eps(mult sub)",down%eps,tag)

    i=srite(tag,0.,4)
    i=sseek(tag,0,0)

    create_image_file=T
  end function create_image_file

  !--------------------------------------------------------------
  !
  !
  !  Core routine to setup an image
  !
  !
  logical function image_setup(X,zero)
    type(weimag), intent(inout) :: X
    integer :: er
    logical :: zero
#ifdef DEBUG
    character(len=128), parameter :: unit='image_setup'
    if(run%debug) call in(unit)
#endif
    image_setup=F
    X%bites=X%esize*&
            product((/down%aix%n,down%aiy%n,down%ahx%n,down%ahy%n,down%nzs/))

    !! DEPTH blocking
    if (.not.down%autonzs) then
    if( (1.0*X%az%n)/down%nzs /= X%az%n/down%nzs ) then
       call image_report(X)
       call seperr('R.az.n is not a multiple of nzs')
    end if
    endif

    down%nzb = down%az%n / down%nzs
    if (.not. down%image_freqs) then
       allocate(X%bin_tot(down%aix%n,down%aiy%n,down%ahx%n,down%ahy%n,down%nzs),stat=er)
       if(er/=0) call seperr('cant allocate R.bin.big')
       allocate(X%bin    (down%amx%n,down%amy%n,down%ahx%n,down%ahy%n,down%nzs),stat=er)
       if(er/=0) call seperr('cant allocate Rbin')
    else
       X%bites=X%esize*&
               product((/down%aix%n,down%aiy%n,down%ahx%n,down%ahy%n,down%nws/))
       allocate(X%bin_tot(down%aix%n,down%aiy%n,down%ahx%n,down%ahy%n,1),stat=er)

       if(er/=0) call seperr('cant allocate R.bin.big')
       allocate(X%bin    (down%amx%n,down%amy%n,down%ahx%n,down%ahy%n,1),stat=er)
       if(er/=0) call seperr('cant allocate Rbin')
    end if
    X%bin_tot=(0.0,0.0)
    X%bin=(0.0,0.0)

    if(run%m_nth==1 .and. zero) then
       call image_assign(X,cmplx(0.,0.))
    end if
    if(run%mpi) then
      if(.not. mpi_image_setup(X,zero)) then
        write(0,*)  "trouble setting up mpi image files"
        return
      end if 
    end if

    image_setup=T
#ifdef DEBUG
    call image_report(X)
    if(run%debug) call out(unit)
#endif
  end function image_setup


    !----------------------------------------------------------------
  !
  !   Setup image structure for local files if
  !     we are running MPI
  !
  !
  logical function setup_lead_image_file(X,lead)
    type(weimag), intent(inout) :: X
    logical :: lead
    integer :: ierr,nsz
    character(len=128) :: first_lock_file,aa

      setup_lead_image_file=.false.

      !MAKE SURE THE FIRST THREAD IS A LEAD
      call MPI_Barrier(MPI_COMM_WORLD,ierr)
      nsz=len(trim(down%image_lock))
      call MPI_Bcast(nsz,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
      first_lock_file=down%image_lock
      call MPI_Bcast(first_lock_file,nsz,MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)


      !FIGURE OUT WHAT THE LEAD TRACE FOR EACH MACHINE
      down%image_lock_fd=-1
      if(first_lock_file(1:nsz)/=down%image_lock(1:nsz) .or. &
        run%m_ith == 0) &
         down%image_lock_fd=mpi_sep_check_lock(down%image_lock)
      lead=.true. 
      if(-1==down%image_lock_fd) lead=.false.


      !CREATE THE LEAD FILE
      if(lead) then
        if(.not. create_image_file(X,X%tag)) then
          write(0,*) "trouble setting up mpi data"
          return
        end if
      end if
      !WAIT UNTIL THE LEAD FILES HAVE BEEN CREATED
      call MPI_Barrier(MPI_COMM_WORLD,ierr)
      !IF WE AREN'T THE LEAD FILE OPEN UP THE IMAGE FILE WITH THE RIGHT DATAPATH
      if(.not.lead) then
           call auxinout(X%tag)
!           call auxclose(X%tag)
       end if
      setup_lead_image_file=.true.
    end function setup_lead_image_file
!----------------------------------------------------------------
  logical function mpi_image_setup(X,zero)
    type(weimag), intent(inout) :: X
    logical :: zero,create
    integer :: ierr,nsz,n,i
    mpi_image_setup=F

    X%tag_0=X%tag
    if(run%share_image) then
       write(X%tag,"(a,a)") trim(X%tag_0),run%mach
    else
!       write(X%tag,"(a,a)") trim(X%tag_0),run%tag_num
    write(X%tag,*) trim(X%tag_0),trim(run%tag_num)
    i=1
    do while(i<128 .and. X%tag(i:i)==" ")
      i=i+1
    end do
    X%tag=X%tag(i:len(trim(X%tag)))
    end if 

    if(run%share_image) then
      if(.not.setup_lead_image_file(X,create)) then
        write(0,*) "trouble setting up lead image file"
        return
      end if
    else
!       X%aw_local%n=down%nw_local*down%nws

       n=X%aw%n
       X%aw%n=down%nw_local*down%nws

       create=.true.
      if(.not. create_image_file(X,X%tag)) then
        write(0,*) "trouble setting up mpi data"
        return
      end if
      X%aw%n=n
    end if 
    if(run%share_image) image_comm_group= create_group(create)
    image_lead_thread=create
    if(create) then
      if(.not. zero) then
        call distribute_image(X,create)
      else 
        call  image_assign(X,cmplx(0.,0.),.true.)
      end if
    end if
    if(create)call mpi_sep_close_lock(down%image_lock,down%image_lock_fd)

    mpi_image_setup=T
  end function mpi_image_setup
  !----------------------------------------------------------------
  !
  !  Assign value to the image file
  !
  subroutine image_assign(X,v,ignore_lock)
    type(weimag), intent(inout) :: X
    complex, intent(in) :: v
    integer :: izb
    logical, optional :: ignore_lock
    logical :: ignore
#ifdef DEBUG
    character(len=128), parameter :: unit='image_assign'
    if(run%debug) call in(unit)
#endif

    ignore=.false.
    if(present(ignore_lock)) ignore=ignore_lock
    X%bin=v
    do izb=1,down%nzb
       call image_write(X,izb,ignore_lock=ignore)
    end do

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine image_assign
  !----------------------------------------------------------------
  !
  ! Write info about image
  subroutine image_report(X)
    type(weimag), intent(in) :: X

    write(0,*)'        *       ',trim(X%tag),' from ',trim(X%from)
    write(0,'(a,i4,f10.2,f10.3)       ') "         *       Ix: ",X%aix%n, X%aix%o, X%aix%d
    write(0,'(a,i4,f10.2,f10.3)       ') "         *       Iy: ",X%aiy%n, X%aiy%o, X%aiy%d
    write(0,'(a,i4,f10.2,f10.3)       ') "         *Wrk areax: ",X%amx%n, X%amx%o, X%amx%d
    write(0,'(a,i4,f10.2,f10.3)       ') "         *Wrk areay: ",X%amy%n, X%amy%o, X%amy%d
    write(0,'(a,i4,f10.2,f10.3)       ') "         *from Shtx: ",down%asx%n, down%asx%o, down%asx%d
    write(0,'(a,i4,f10.2,f10.3)       ') "         *from Shty: ",down%asy%n, down%asy%o, down%asy%d
    write(0,'(a,i4,f10.2,f10.3)       ') "         *     CIGx: ",X%ahx%n, X%ahx%o, X%ahx%d
    write(0,'(a,i4,f10.2,f10.3)       ') "         *     CIgy: ",X%ahy%n, X%ahy%o, X%ahy%d
    write(0,'(a,i4,f10.2,f10.3,a,i3,a)') "         *        Z: ",X%az%n,  X%az%o,  X%az%d," ->",down%nzb," blocks"
    if(down%image_freqs)&
    write(0,'(a,i4,f10.2,f10.3)       ') "         *    freqs: ",X%aw%n, X%aw%o, X%aw%d
    if(associated(X%bin)) then
    write(0,'(a,f9.2,a,f9.2,a)        ') "         *   block: ",X%bites/1024./1024.," / ",X%esize*size(X%bin)/1024./1024.," Mbytes"
    write(0,'(a,i1)')                    "         *   esize= ",X%esize
    end if
    write(0,*)'        ************************************************'
  end subroutine image_report
!----------------------------------------------------------------
!
! Function image_insert: Places single-shot image subspace into 
!                         total image space. By brad 11/02/03
logical  function image_insert(X)
    type(weimag), intent(in)   ::X

    image_insert=F

    X%indx= (1.5+ (down%amx%o-down%aix%o)/down%amx%d)
    X%indx= min(X%indx,down%aix%n-down%amx%n) !! Don't go off min/max points
    X%indx= max(X%indx,1)

    X%indy= (1.5+ (down%amy%o-down%aiy%o)/down%amy%d)
    X%indy= min(X%indy,down%aiy%n-down%amy%n) !! Don't go off min/max points
    X%indy= max(X%indy,1)

    X%finx=X%indx + down%amx%n -1
    X%finy=X%indy + down%amy%n -1
    if(.not.down%image_freqs) then
       X%bin_tot(X%indx:X%finx,X%indy:X%finy,:,:,:) = &
       X%bin_tot(X%indx:X%finx,X%indy:X%finy,:,:,:) + X%bin
    else
       X%bin_tot(X%indx:X%finx,X%indy:X%finy,:,:,:) = X%bin
    end if
    X%bin=(0.,0.)

    image_insert=T
  end function image_insert
!----------------------------------------------------------------
  subroutine Rread(X,k)
    type(weimag), intent(inout) :: X
    integer,      intent(in)    :: k
    integer :: stat
#ifdef DEBUG
    character(len=128), parameter :: unit='weiR_read'
!    if(wpro%dbg>1) call in(unit)
    call in(unit)
#endif 
!    down%image_lock_fd= mpi_sep_open_lock(down%image_lock)
    
    stat=sseek_block(X%tag,k-1,X%bites,0)
    if(X%bites /= sreed(trim(X%tag), X%bin, X%bites)) &
      call seperr('trouble in subroutine Rread(X,k)',X%tag)

!    call mpi_sep_close_lock(down%image_lock,down%image_lock_fd)

#ifdef DEBUG
    call out(unit)
!    if(wpro%dbg>1) call out(unit)
#endif
  end subroutine Rread

  !----------------------------------------------------------------
  !
  !
  ! Function to read image
  !
  !
  subroutine image_read(X,k,iw)
    type(weimag), intent(inout) :: X
    integer, optional :: iw
    integer, intent(in) :: k
    integer :: stat,i_w,i4,i5,i3,ierr;
    real, allocatable :: temp_buf(:,:)
#ifdef DEBUG
    character(len=128), parameter :: unit='image_read'
    if(run%debug) call in(unit)
#endif
      down%image_lock_fd= mpi_sep_open_lock(down%image_lock)

    i_w=0; if(present(iw)) i_w=iw
    if(i_w==0) then 
      stat=sseek_block(X%tag,k-1,X%bites,0)
      if(X%esize==4) then
         allocate(temp_buf(size(X%bin_tot,1),size(X%bin_tot,2)), stat=stat)
         do i5=1,size(X%bin_tot,5)
         do i4=1,size(X%bin_tot,4)
         do i3=1,size(X%bin_tot,3)
            if(size(temp_buf)*4/=sreed(trim(X%tag), temp_buf,size(temp_buf)*4)) then
              write(0,*) "trouble reading image block",i3,i4,i5,iw
              return
            end if 
            X%bin_tot(:,:,i3,i4,i5)=cmplx(temp_buf,0.)
         end do; end do; end do;
         deallocate(temp_buf)
      else
         ierr= sreed(trim(X%tag), X%bin_tot, X%bites)
      end if
    else
      call seperr("unimplented, varaible sampling image read")
    end if
      call mpi_sep_close_lock(down%image_lock,down%image_lock_fd)

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine image_read
  !----------------------------------------------------------------
  !
  !  Function to combine image
  subroutine combine_image(X)
    type(weimag), intent(inout) :: X
    integer ::  ierr,iadd,i
    integer, external :: sseek
    
#ifdef SEP_MPI
    call MPI_Barrier(MPI_COMM_WORLD,ierr)
    iadd=0;if(down%add) iadd=1;
    if(.not.down%image_freqs)then
       if(run%mpi .and. down%adj .and. image_lead_thread) then
          ierr=sseek(trim(X%tag),0,0) 
          if(run%m_ith==0) ierr= sseek(trim(x%tag_0),0,0)
          if(0/=mpi_sep_tag_sum(run%m_ith,X%tag,0,X%tag_0,X%bites/4,&
          down%nzb,iadd)) call seperr("trouble combining image ");
       end if
       call MPI_Barrier(MPI_COMM_WORLD,ierr)
       call aux_unlink(X%tag)
       if(var%use_disk_buf) call aux_unlink(X%tag_var)
    else
                        ierr=sseek(trim(X%tag)  ,0,0)
       if(run%m_ith==0) ierr=sseek(trim(x%tag_0),0,0)
       do i=1,X%az%n
          if(0/=mpi_sep_tag_combine(run%m_ith,X%tag,0,X%tag_0, &
          X%bites/4,down%nwp,down%iw_thread,iadd)) &
          call seperr("trouble combining image")
       end do
       call aux_unlink(X%tag)
    end if
#endif
  end subroutine combine_image

  !----------------------------------------------------------------
  !
  !  Distribute the image to various threads
  subroutine distribute_image(X,use)
    type(weimag), intent(inout) :: X
    integer :: ierr,iuse,i
    logical :: use

#ifdef SEP_MPI
    if(run%mpi) then
        ierr= sseek(trim(X%tag),0,0)
       if(run%m_ith==0) then
          ierr=sseek(trim(X%tag_0),0,0)
        end if
       if(run%share_image) then
          i=1; do while(image_master(i)==0);i=i+1;end do;i=i-1;
         if(0/=mpi_sep_tag_bcast_index(0,X%tag_0,run%m_ith,X%tag,X%bites/4,&
           down%nzb,image_comm_group))call seperr("trouble distributing image 1")
       else
         if(0/=mpi_sep_tag_bcast(0,X%tag_0,run%m_ith,X%tag,X%bites/4,&
           down%nzb)) call seperr("trouble distributing image 2")
       end if
    end if
#endif
  end subroutine distribute_image
!------------------------------------------------------------------------
  integer function create_group(use)
   logical ::use
   integer :: iuse,i
   integer :: ierr
   integer :: status(MPI_STATUS_SIZE)

   if(use) then
     iuse=1
   else 
     iuse=0
   end if

#ifdef SEP_MPI
   allocate(image_master(run%m_nth))
   do i=1,run%m_nth 
     if(run%m_ith==0 .and. i/=1) then
        call mpi_recv(image_master(i),1,MPI_INTEGER,i-1,3242+i,MPI_COMM_WORLD,status,ierr);
     else if(i-1 == run%m_ith .and. run%m_ith/=0) then
        call mpi_send(iuse,1,MPI_INTEGER,0,3242+i,MPI_COMM_WORLD,ierr)
     end if
   end do
   if(run%m_ith==0) image_master(1)=iuse
!   call MPI_Barrier(MPI_COMM_WORLD)
   call  mpi_bcast(image_master,size(image_master),MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
   create_group= mpi_sep_create_group(size(image_master),image_master)
#endif
 end function create_group
  !----------------------------------------------------------------
  !
  !  Write out image
  !
  subroutine image_write(X,k,sum,iw,nz,ignore_lock)
    logical, optional :: sum
    logical           :: sum_it
    integer, optional :: iw,nz
    integer           :: i_w,n_z,i3,i4,ierr
    type(weimag), intent(inout) :: X
    integer, intent(in) :: k
    integer :: iz,i,izlast,izfirst,ib,stat,n123
    logical,optional :: ignore_lock
    logical :: ignore
    real    :: z,zcur,ab
    logical :: write_out,stored
    complex, allocatable :: tempc_buf(:,:)
    complex, allocatable :: temp2_buf(:,:),temp3_buf(:,:,:)
#ifdef DEBUGx
    character(len=128), parameter :: unit='image_write'
    if(run%debug) call in(unit)
#endif

!    if(image_lead_thread .and. .not. ignore_lock) return
!    if(.not. image_lead_thread) return
    ignore=.false.;if(present(ignore_lock)) ignore=ignore_lock
    if(run%share_image .and. .not. ignore)  &
      down%image_lock_fd= mpi_sep_open_lock(down%image_lock)
!    if(run%share_image) call auxinout(X%tag)
    sum_it=.false.
    if(present(sum)) sum_it=sum
    i_w=0; if(present(iw)) i_w=iw
    n_z=0; if(present(nz)) n_z=nz

    n123=size(X%bin_tot(:,:,:,:,i))

!    if(i_w==0) then      !Just doing a normal write
      iz=(k-1)*size(X%bin_tot,5)
      do i=1,size(X%bin_tot,5)
        iz=iz+1
        call image_write_iz(X,X%bin_tot(:,:,:,:,i),iz,sum_it)
      end do
!    else
!      allocate(tempc_buf(size(X%bin_tot,1),size(X%bin_tot,2)))
!      allocate(temp2_buf(size(X%bin_tot,1),size(X%bin_tot,2)))
!      allocate(temp3_buf(size(X%bin_tot,1),size(X%bin_tot,2),size(X%bin_tot,5)))
!
!      !
!      !
!      !First check to see what we can write
!      !
!      !
!
!      izlast=min(size(var%z_samp(i_w)%dz),size(X%bin_tot,5)*k) !last iz step done
!      zcur=var%z_samp(i_w)%zloc(izlast) ! depth we have gone done to
!      z=down%az%o;iz=1;
!      do while(z<=zcur .and. iz <= down%az%n) 
!        if(.not. var%filled(iz)) then
!          write_out=.true.
!          do i=1,size(var%z_samp(i_w)%iloc,1) 
!            if(var%z_samp(i_w)%iloc(i,iz) > izlast .and.  &
!               abs(var%z_samp(i_w)%frac(i,iz)) > .001) &
!                write_out=.false.
!          end do 
!          if(write_out) then
!            stat=sseek_block(X%tag,iz-1,X%bites/size(X%bin_tot,5),0)
!            do i4=1,size(X%bin_tot,4)
!              do i3=1,size(X%bin_tot,3) 
!                tempc_buf=0.
!                temp3_buf=X%bin_tot(:,:,i3,i4,:)
!                 ab=maxval(real(temp3_buf))
!!                write(0,*) "IN LAGRANGE",run%m_ith,iz,i3,i4,maxval(real(temp3_buf))
!                call lagrange_interp(.true.,i_w,k,iz,temp3_buf,tempc_buf,&
!                   temp2_buf,X%tag_var,i3,i4)
!                call write_mx_my(X%tag,X%esize,size(X%bin_tot,1),size(X%bin_tot,2)&
!                  ,tempc_buf,sum_it)
!              end do
!            end do
!            var%filled(iz)=.true.
!          end if 
!         end if
!         z=z+down%az%d;iz=iz+1
!       end do
!
!      !
!      !
!      ! Check to see if we can clear any incore or disk memory
!      !
!
!      ! Find the smallest iz still needed
!      izfirst=izlast
!      do iz=1, size(var%filled) 
!        if(.not. var%filled(iz)) then
!          do i=1,size(var%z_samp(i_w)%iloc,1)
!            if(abs(var%z_samp(i_w)%frac(i,iz))>.001) &
!              izfirst=min(izfirst,var%z_samp(i_w)%iloc(i,iz)) 
!          end do
!        end if 
!      end do 
!
!      do i=1,var%ndisk_buf
!        if(var%disk_buf(i) < izfirst ) var%disk_buf(i)=0
!      end do
!    
!      do i=1,var%nmem_buf
!        if(var%mem_buf(i) < izfirst ) var%mem_buf(i)=0
!      end do
!    
!      ! 
!      ! 
!      !store any newly created image components that are still needed
!      ! 
!      ! 
!
!      iz=(k-1)*size(X%bin_tot,5)
!      do i=1,n_z
!        iz=iz+1
!        if(iz>=izfirst) then !we need to store this image component
!          stored=.false.;ib=1;
!          do while(ib <= var%nmem_buf .and. .not. stored) 
!            if(var%mem_buf(ib)==0) then
!              var%v_i_buf(:,:,:,:,ib)=X%bin_tot(:,:,:,:,i)
!              stored=.true.; var%mem_buf(ib)=iz
!              var%z_samp(iw)%status(iz)=ib
!            end if 
!            ib=ib+1; 
!          end do
!          ib=1;
!          do while(ib <= var%ndisk_buf .and. .not. stored) 
!            if(var%disk_buf(ib)==0) then
!              stat=sseek_block(X%tag_var,ib-1,n123*8,0)
!              ierr= srite(X%tag_var, X%bin_tot(:,:,:,:,i), n123*8)
!              stored=.true.; var%disk_buf(ib)=iz
!              var%z_samp(iw)%status(iz)=-ib
!            end if 
!            ib=ib+1;
!          end do
!        end if 
!      end do
!      deallocate(tempc_buf,temp2_buf,temp3_buf)
!    end if

!    if(run%share_image) call auxclose(X%tag)
!    if(run%share_image .and. .not. ignore) &
!      call mpi_sep_close_lock(down%image_lock,down%image_lock_fd)
  end subroutine image_write
!--------------------------------------------------------------
  subroutine image_write_iz(X,image,k,sum_it)
    logical           :: sum_it
    type(weimag), intent(inout) :: X
    integer, intent(in) :: k
    integer :: stat,i3,i4,blk,rem
    complex :: image(:,:,:,:)
     
#ifdef DEBUG
    character(len=128), parameter :: unit='image_write_iz'
    if(run%debug) call in(unit)
#endif

       stat=sseek_block(X%tag,k-1,X%bites/size(X%bin_tot,5),0)
       do i4=1,size(X%bin_tot,4)
         do i3=1,size(X%bin_tot,3)
             call write_mx_my(X%tag,X%esize,size(image,1),size(image,2),image(:,:,i3,i4),sum_it)
          end do
       end do
!      call to_history("zblock",k,X%tag)
   
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine image_write_iz
!--------------------------------------------------------------
  subroutine write_mx_my(tag,esize,n1,n2,image,sum)
    logical :: sum
    character(len=*) :: tag
    integer :: n1,n2,ierr,esize,blk,rem
    complex :: image(:,:),tempc(n1,n2)
    real :: tempr(n1,n2)

    if(esize==4) then
      if(sum) then
        if(esize*n1*n2/=sreed(tag,tempr,n1*n2*4)) &
          call erexit("trouble reading in image 4")
        tempr=tempr+real(image) 
        ierr=sseek(tag,-n1*n2*4,1)
      else 
        tempr=real(image)
      end if
      if(esize*n1*n2/=srite(tag,tempr,n1*n2*4)) &
        call erexit("trouble writing out image 4")
    else

      if(sum) then
        if(esize*n1*n2/=sreed(tag,tempc,n1*n2*8)) &
          call erexit("trouble reading in image 8")
        tempc=tempc+image
        ierr=sseek(tag,-n1*n2*8,1)
      else
        tempc=image
      end if
      if(8*n1*n2/=srite(tag,tempc,n1*n2*8)) &
        call erexit("trouble writing out image 8" )
    end if
  end subroutine write_mx_my
  !----------------------------------------------------------------
  !
  !
  ! Image initializing routine when the image file exists
  !
  !
  logical function image_history_init(X)
    type(weimag), intent(inout) :: X

#ifdef DEBUG
    character(len=128), parameter :: unit='image_history_init'
    if(run%debug) call in(unit)
#endif
    image_history_init=F

    call auxinout(X%tag)
    X%from = "history"

    call from_aux(X%tag,"esize",X%esize,4)

    call hinaxis(X%tag,X%aix,1); X%aix%l='ix'

    call hinaxis(X%tag,X%aiy,2); X%aiy%l='iy'

    call hinaxis(X%tag,X%az ,5); X%az%l ='z_'

    call hinaxis(X%tag,X%ahx,3); X%ahx%l='hx'
      
    call hinaxis(X%tag,X%ahy,4); X%ahy%l='hy'

    if(.not. image_setup(X,.false.)) then
       write(0,*) "trouble setting up image"
       return
    end if

    image_history_init=T

#ifdef DEBUG
    if(run%debug) call image_report(X)
    if(run%debug) call out(unit)
#endif
  end function image_history_init
  !----------------------------------------------------------------
end module wei_obj_R

