module wei_obj_S

  use wei_run
  use wei_down_type
  use wei_var
  use wei_var_samp
  use wei_type
  !  use wei_work
  use wei_xmap
  !---------------
  implicit none

  real,    dimension(:), pointer, private ::  sloline
  type(weslowmap), public :: slo
  type(weslow),public :: S

contains
  !----------------------------------------------------------------
  !INITIALIZE THE SLOWNESS FUNCTION
  ! GIVEN UP TO TWO TAGS DESCRIBING THE SLOWNESS
  logical function wei_slow_init(X,tag_S,tag_G)
    type(weslow),       intent(inout) :: X
    character(len=128), intent(in)    :: tag_S,tag_G
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='wei_slow_init' 
    if(run%debug) call in(unit)
#endif


    wei_slow_init=F 

    if(down%conv_wave) then
       X%s_s=X%s1
       X%s_g=X%s2
       X%s_s%tag = tag_S
       X%s_g%tag = tag_g
       if(.not. slowness_history_init(X%s_s)) then
          write(0,*) "trouble init slowness history"
          return
       end if
       if(.not. slowness_history_init(X%s_g)) then
          write(0,*) "trouble init slowness history"
          return
       end if
    else
       X%s_s=X%s1
       X%s_g=X%s1
       X%s_s%tag = tag_S
       if(.not. slowness_history_init(X%s_s)) then
          write(0,*) "trouble init slowness history"
          return
       end if
    end if

    call slow_unitcheck(S%s_s)

    !allocate array
    allocate(X%pba(down%amx%n,down%amy%n,down%nzs),stat=er)
    if(er/=0) call seperr('cant allocate S.pba')
    X%pba=(0.0,0.0)

    call find_max_slow()

    down%S=S
    wei_slow_init=T 

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function wei_slow_init




  !----------------------------------------------------------------
  !
  !INITIALIZE SLOWNESS FROM A FILE
  !
  logical function slowness_history_init(X)
    type(weslow_comp), intent(inout) :: X
    integer  :: stat
    character(len=128) :: temp
#ifdef DEBUG
    character(len=128), parameter :: unit='slowness_history_init' 
    if(run%debug) call in(unit)
#endif

    slowness_history_init=F
!    call auxinout(X%tag)
    X%from = "history"

    call hinaxis(X%tag,X%amx,1); X%amx%l='ix'
    call hinaxis(X%tag,X%amy,2); X%amy%l='iy'
    call hinaxis(X%tag,X%az ,3); X%az%l ='z_'
    down%az = X%az;down%bz = down%az
    if (.not.down%autonzs) down%bz%n=down%nzs

    call from_param("velocity",X%velocity,.false.)
    call from_aux(trim(X%tag),"esize",X%esize,4)

    if(.not. slowness_setup(X)) then
       write(0,*) "trouble with slowness setup"
       return
    end if

    if(X%tag_0(1:1)=="/") then
       temp=""
    else
       temp=trim(run%local_path)
    end if

    if(down%order(1:3)=="VAR") then
       write(X%tag_var,"(a,a,a,a)") trim(temp),trim(X%tag_0),"VS_",run%tag_num
       call auxscr(X%tag_var)
       stat= srite(X%tag_var,0.,4)
    end if
    slowness_history_init=T

#ifdef DEBUG
    if(run%debug) call slowness_report(X)
    if(run%debug) call out(unit)
#endif
  end function slowness_history_init
!----------------------------------------------------------------
! . . A few sanity checks
subroutine slow_unitcheck(X)
  type(weslow_comp)::X
  real:: sloc,loc
!! . . Sanity checks
  if (.not.down%passive) then
    if (X%amx%o .gt. down%aix%o )&
     call seperr("Slowness model origin greater than data coverage: x-axis")
    if (X%amy%o .gt.down%aiy%o )&
     call seperr("Slowness model origin greater than data coverage: y-axis")
    if (((down%amx%d/X%amx%d .gt. 100).or.(down%amx%d/X%amx%d .lt. 0.01 )).and.(X%amx%n /=1))&
     call seperr("check units are of same magnitude between slowness data files: x-axis")
    if (((down%amy%d/X%amy%d .gt. 100).or.(down%amy%d/X%amy%d .lt. 0.01 )).and.(X%amy%n /=1)) &
     call seperr("check units are of same magnitude between slowness data files: y-axis")

    if(down%marine)then
!if(run%verb)write(0,*) down%asx%o,down%asx%n,down%asx%d ,down%amx%o,down%amx%n,down%amx%d
!if(run%verb)write(0,*)X%amx%o,X%amx%n,X%amx%d
       loc=(down%asx%n-1)*down%asx%d + down%amx%o+(down%amx%n-1)*down%amx%d
    else
!if(run%verb)write(0,*) down%asx%o,down%asx%n,down%asx%d ,down%amx%o,down%amx%n,down%amx%d
       loc=down%asx%o+(down%asx%n-1)*down%asx%d + down%amx%o+(down%amx%n-1)*down%amx%d
!if(run%verb)write(0,*) 'loc',loc
    end if
       sloc=X%amx%o + (X%amx%n-1)*X%amx%d
!if(run%verb)write(0,*) 'sloc',sloc
!    if((loc .gt. sloc) .and. (down%amx%n .gt. 1)) &
!     call seperr("the slowness model is not long enough: x-axis")
!    loc=(down%asy%n-1)*down%asy%d + down%amy%o+(down%amy%n-1)*down%amy%d
!    sloc=X%amy%o + (X%amy%n-1)*X%amy%d
!    if((loc .gt. sloc).and. (down%amy%n .gt. 1)) &
!     call seperr("the slowness model is not long enough: y-axis")
! end if
end if
  end subroutine slow_unitcheck

  !----------------------------------------------------------------
  ! SETUP SLOWNESS
  ! 
  logical function slowness_setup(X)
    type(weslow_comp), intent(inout) :: X
    integer :: er
    character(len=128) :: temp
#ifdef DEBUG
    character(len=128), parameter :: unit='slowness_setup' 
    if(run%debug) call in(unit)
#endif

    slowness_setup=F
    X%bites = X%esize * product( (/X%amx%n,X%amy%n,down%nzs/) )

#ifdef SEP_MPI
    if(X%tag_0(1:1)=="/") then
       temp=""
    else
       temp=trim(run%local_path)
    end if

    if(run%mpi) then
       X%tag_0=X%tag;
       write(X%tag_local,"(a,a,a,a)") trim(temp),"brad",trim(X%tag_0),run%tag_num
       x%tag=X%tag_local
!       call auxinout(trim(X%tag))
       call auxscr(trim(X%tag))
       call to_history("esize",X%esize,trim(X%tag))
       call houaxis(trim(X%tag),X%amx,1)
       call houaxis(trim(X%tag),X%amy,2)
       call houaxis(trim(X%tag),X%az, 3)
       call to_history("esize",X%esize,trim(X%tag))
    end if
    call distribute_slow(X)
#endif

    if( (1.0*X%az%n)/down%nzs /= X%az%n/down%nzs ) then
       call slowness_report(X)
       call seperr('S.az.n is not a multiple of nzs')
    end if
  
    down%nzb = down%az%n / down%nzs
    !!--------------------
    allocate(X%bin(X%amx%n,X%amy%n,down%nzs),stat=er)
    if(er/=0) call seperr('cant allocate S.bin')
    X%bin=(0.0,0.0)

    !!--------------------
    !! slowness interpolation map
    allocate( sloline(down%nzs),stat=er)
    if(er/=0) call seperr('cant allocate  sloline')
    slowness_setup=T
    if(run%verb) call slowness_report(X)
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function slowness_setup
!----------------------------------------------------------------
  !FIND THE MAXIMUM SLOWNESS AT EACH DEPTHS
  subroutine find_max_slow()
    integer :: i,ndone,block,j,ierr
    character(len=128) :: tag_1,tag_2
    real,allocatable :: junk(:)
    ndone=0;
    allocate(junk(S%s_s%az%n))
    allocate(down%max_slow(S%s_s%az%n))
    i=0
    do while(ndone< S%s_s%az%n)
       i=i+1
       block=min(s%s_s%az%n-ndone,size(S%s_s%bin,3))
       call slowness_read(S,i,block)
       if(down%conv_wave) then
          do j=1,block
             down%max_slow(j+ndone)=max(maxval(real(S%s_s%bin(:,:,j))),maxval(real(S%s_g%bin(:,:,j))))
          end do
       else
          do j=1,block
             down%max_slow(j+ndone)=maxval(real(S%s_s%bin(:,:,j)))
          end do
       end if
       ndone=ndone+block
    end do
    deallocate(junk)
  end subroutine find_max_slow

  !----------------------------------------------------------------
  !WRITE INFO ABOUT SLOWNESS
  subroutine slowness_report(X)
    type(weslow_comp), intent(in) :: X
    write(0,*)'        *       ',trim(X%tag),' from ',trim(X%from)
    write(0,'(a,i4,f10.2,f10.3)       ') "         *    Mx: ",X%amx%n, X%amx%o, X%amx%d
    write(0,'(a,i4,f10.2,f10.3)       ') "         *    My: ",X%amy%n, X%amy%o, X%amy%d
    write(0,'(a,i4,f10.2,f10.3,a,i3,a)') "         *     Z: ",X%az%n,  X%az%o,  X%az%d," ->",down%nzb," blocks"
    if(X%velocity) then
       write(0,'(a,i1,a)')               "         *    esize=",X%esize,"                 VELOCITY"
    else 
       write(0,'(a,i1,a)')               "         *    esize=",X%esize,"                 SLOWNESS"
    end if
    if(associated(X%bin)) then
       write(0,'(a,f9.2,a,f9.2,a)        ') "         * block: ",X%bites/1024./1024.," / ",8*size(X%bin)/1024./1024.," Mbytes"
       write(0,*)'        ************************************************'
    end if

  end subroutine slowness_report
  !--------------------------------------------------------------
  ! READ SLOWNESS
  subroutine slowness_read(X,k,nzs)
    type(weslow), intent(inout) :: X
    integer, intent(in) :: k
    integer, optional :: nzs
    integer :: stat,nz,i
    real,allocatable  :: temp_buf(:,:,:)
#ifdef DEBUG
    character(len=128), parameter :: unit='slowness_read' 
    if(run%debug) call in(unit)
#endif

    call slowness_read_comp(X%s_s,k,nzs)
    if(down%conv_wave) then
       call slowness_read_comp(X%s_g,k,nzs)
    end if

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine slowness_read
  !----------------------------------------------------------------
  !READ A SLOWNESS COMPONENT
  subroutine slowness_read_comp(X,k,nzs)
    type(weslow_comp), intent(inout) :: X
    integer, intent(in) :: k
    integer, optional :: nzs
    integer :: stat,nz,i
    real,allocatable  :: temp_buf(:,:,:)
#ifdef DEBUG
    character(len=128), parameter :: unit='slowness_read_comp' 
    if(run%debug) call in(unit)
#endif

    nz=size(X%bin,3); if(present(nzs)) nz=nzs
    stat=sseek_block(X%tag,k-1,X%bites,0)
    if(X%esize==4) then
       allocate(temp_buf(size(X%bin,1),size(X%bin,2),size(X%bin,3)),stat=stat)
       if(stat/=0) call erexit("trouble allocating real temp slow buf")
       stat= sreed(X%tag, temp_buf, X%bites/size(X%bin,3)*nz)
       if(X%velocity) temp_buf=1./temp_buf
       X%bin=cmplx(temp_buf,temp_buf)
       deallocate(temp_buf)
    else
       stat= sreed(X%tag, X%bin, X%bites)
    end if

    if(nz <  size(X%bin,3)) then
       do i=nz,size(X%bin,3)
          X%bin(:,:,i)=X%bin(:,:,nz)
       end do
    end if

#ifdef DEBUG 
    if(run%debug) call out(unit)
#endif
  end subroutine slowness_read_comp

  !--------------------------------------------------------------
  !DISTRIBUTE SLOWNESS TO ALL THE THREADS
  subroutine distribute_slow(X)
    integer :: stat
    type(weslow_comp), intent(inout) :: X

#ifdef SEP_MPI
    if(run%mpi) then
       stat= sseek(trim(X%tag),0,0)
       if(run%m_ith==0) then
          stat= sseek(trim(X%tag_0),0,0);
       end if
       if(0/=mpi_sep_tag_bcast(0,X%tag_0,run%m_ith,X%tag,X%bites/4,&
       X%az%n/down%nzs)) call seperr("trouble distributing slow ")
    end if
#endif
  end subroutine distribute_slow
  !--------------------------------------------------------------
  !----------------------------------------------------------------
  ! WRITE OUT SLOWNESS
  subroutine slowness_write(X,k)
    type(weslow_comp), intent(inout) :: X
    integer :: k
    integer :: stat
    integer :: iz
#ifdef DEBUG
    character(len=128), parameter :: unit='slowness_write'
    if(run%debug) call in(unit)
#endif

    stat=sseek_block(X%tag,k-1,X%bites,0)
    stat= srite(X%tag, X%bin, X%bites)

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine slowness_write
  !----------------------------------------------------------------




  !ALLOCATE SPACE FOR SLOWNESS
  subroutine weislow_allocate()
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weislow_allocate' 
    if(run%debug) call in(unit)
#endif

    !! FKop slowness arrays
    allocate(slo%sz_r(down%nzs,down%nref+1),stat=er)
    if(er/=0) call seperr('cant allocate slo%sz_r')
    allocate(slo%sz_s(down%nzs,down%nref+1),stat=er)
    if(er/=0) call seperr('cant allocate slo%sz_s')

    allocate(slo%nrs_s(down%nzs),stat=er)
    if(er/=0) call seperr('cant allocate slo%nrs_nzs')
    allocate(slo%nrs_r(down%nzs),stat=er)
    if(er/=0) call seperr('cant allocate slo%nrs_g')

    allocate(slo%bound_r(down%nzs,down%nref),stat=er)
    if(er/=0) call seperr('cant allocate slo%bound_r')
    allocate(slo%bound_s(down%nzs,down%nref),stat=er)
    if(er/=0) call seperr('cant allocate slo%bound_s')

    !! FXop interpolation array
    allocate(slo%si_r(down%amx%n,down%amy%n,down%nzs),stat=er)   
    if(er/=0) call seperr('cant allocate slo%si_r')
    allocate(slo%si_s(down%amx%n,down%amy%n,down%nzs),stat=er)   
    if(er/=0) call seperr('cant allocate slo%si_s')

!    allocate(slo%th(run%o_nth))
    allocate(slo%th(1))
    slo%th(1)%si=>slo%si_s
    slo%th(1)%bound=>slo%bound_s
    slo%th(1)%nrs=>slo%nrs_s
    slo%th(1)%sz=>slo%sz_s

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine weislow_allocate
  !----------------------------------------------------------------
  !FIND THE MEAN VALUE OF THE SLOWNESS
  !
  real function slowness_mean(X,k) result(smean)
    type(weslow), intent(in) :: X
    integer :: k

    if(down%conv_wave) then
       smean = (sum(real(X%s_s%bin(:,:,k)))+&
       sum(real(X%s_g%bin(:,:,k))))/(2* X%s_s%amx%n * X%s_s%amy%n)
    else
       smean = sum(real(X%s_s%bin(:,:,k)))/(X%s_s%amx%n * X%s_s%amy%n)
    end if
  end function slowness_mean
  !----------------------------------------------------------------
  !FIND THE MAX VALUE OF THE SLOWNESS
  !
  real function slowness_max(X,k) result(smax)
    type(weslow), intent(in) :: X
    integer :: k

    if(down%conv_wave) then
       smax =max(maxval(real(X%s_s%bin(:,:,k))),maxval(real(X%s_g%bin(:,:,k))))
    else
       smax = maxval(real(X%s_s%bin(:,:,k)))
    end if
  end function slowness_max
  !----------------------------------------------------------------
  !FIND THE MIN VALUE OF THE SLOWNESS
  !
  real function slowness_min(X,k) result(smin)
    type(weslow), intent(in) :: X
    integer :: k

    if(down%conv_wave) then
       smin =min(minval(real(X%s_s%bin(:,:,k))),minval(real(X%s_g%bin(:,:,k))))
    else
       smin = minval(real(X%s_s%bin(:,:,k)))
    end if
  end function slowness_min
  !----------------------------------------------------------------
!  !! push slowness from bin to ss and sr
!  subroutine partition_slow(wfld,ifk,izs,ith)
!    complex :: wfld(:,:,:,:)
!    integer :: imx,imy,ihx,ihy,ifk,i,izs,ith
!    real    :: xc,yc,avg
!    logical :: source
!#ifdef DEBUG
!    character(len=128), parameter :: unit='partition_slow' 
!    if(run%debug) call in(unit)
!#endif
!    do imy=1,down%amy%n
!       do imx=1,down%amx%n
!          if(slo%th(ith)%source) then
!             avg=real(S%pba(imx,imy,izs))
!             i=1; do while(avg > slo%th(ith)%bound(izs,i));i=i+1; end do
!          else
!             avg=aimag(S%pba(imx,imy,izs))
!             i=1; do while(avg > slo%th(ith)%bound(izs,i));i=i+1; end do
!          end if
!          if(i/=ifk) wfld(imx,imy)=cmplx(0.,0.)
!       end do
!    end do
!
!#ifdef DEBUG
!    if(run%debug) call out(unit)
!#endif
!  end subroutine partition_slow
  !----------------------------------------------------------------
  !! push slowness from bin to ss and sr
  subroutine weislow_push(X)
    type(weslow), intent(inout) :: X
    integer :: imx,imy
    real    :: xc,yc
#ifdef DEBUG
    character(len=128), parameter :: unit='weislow_push' 
    if(run%debug) call in(unit)
#endif
    X%pba=(0.,0.)
    if ( (X%s_s%amx%d/=down%amx%d)&
    .and.(X%s_s%amy%d/=down%amy%d))then
       do imy=1,down%amy%n
          do imx=1,down%amx%n
             !! X%s = f(X%bin,x,y)
             xc=xmap%mx(imx)
             yc=xmap%my(imy)
             !! X%s = f(X%bin)
             call weislow_interp (.false.,X%s_s,sloline,xc,yc)
             X%pba(imx,imy,:) = X%pba(imx,imy,:) + (1.,0.)*sloline
             if (down%conv_wave)call weislow_interp(.false.,X%s_g,sloline,xc,yc)
             X%pba(imx,imy,:) = X%pba(imx,imy,:) + (0.,1.)*sloline
          end do
       end do
    else !! just go get the right value, no need to interpolate
       X%indx=1.5 + (down%amx%o - X%s_s%amx%o)/down%amx%d
       X%indy=1.5 + (down%amy%o - X%s_s%amy%o)/down%amy%d
       X%finx=X%indx+down%amx%n-1
       X%finy=X%indy+down%amy%n-1
       X%pba=X%s_s%bin(X%indx:X%finx,X%indy:X%finy,:)
       if (down%conv_wave) then
       X%pba= X%pba + (0.,1.)*X%s_g%bin(X%indx:X%finx,X%indy:X%finy,:)
       X%pba= X%pba + (1.,0.)*X%s_s%bin(X%indx:X%finx,X%indy:X%finy,:)
       end if
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine weislow_push

  !----------------------------------------------------------------
  subroutine weislow_interp(adj,X,line,xc,yc)
    type(weslow_comp), intent(inout) :: X
    logical,                   intent(in) :: adj
    complex, dimension(:,:,:), pointer    :: cube
    real,    dimension(:),     pointer    :: line
    real,                      intent(in) :: xc,yc
    real    :: dx,dy
    integer :: ix,iy, jx,jy
    real    :: fx,fy
    real    :: w00,w01,w10,w11, sumw
#ifdef DEBUG
    character(len=128), parameter :: unit='weislow_interp'
    if(run%debug) call in(unit)
#endif

    cube=>X%bin

    dx = xc
    dy = yc

    !! don't allow coordinates outside the slowness grid
    dx = min(dx,X%amx%o + (X%amx%n-1)* X%amx%d)
    dx = max(dx,X%amx%o)

    dy = min(dy,X%amy%o + (X%amy%n-1)* X%amy%d)
    dy = max(dy,X%amy%o)

    !! lower index
    ix = ( 1.5+ (dx-X%amx%o) / X%amx%d )
    iy = ( 1.5+ (dy-X%amy%o) / X%amy%d )

    !! keep indices in bounds
    ix = min(ix,X%amx%n)
    iy = min(iy,X%amy%n)
    ix = max(ix,        1  ); jx=min(ix+1,X%amx%n)
    iy = max(iy,        1  ); jy=min(iy+1,X%amy%n)

    !! linear weights
    fx = (dx - (X%amx%o + (ix-1)*X%amx%d) ) / X%amx%d
    fy = (dy - (X%amy%o + (iy-1)*X%amy%d) ) / X%amy%d
    fx = max(0., fx)
    fy = max(0., fy)

    !! 00   10       --> x
    !!    .         |
    !! 01   11      v
    !!               y
    w00=(1.-fx)*(1.-fy)
    w10=(   fx)*(1.-fy)
    w01=(1.-fx)*(   fy)
    w11=(   fy)*(   fy)

    !! normalize the sum of the weights
    sumw = w00+w10+w01+w11
    w00 = w00 / sumw
    w10 = w10 / sumw
    w01 = w01 / sumw
    w11 = w11 / sumw

    if(adj) then !! push
       cube(ix,iy,:) = cube(ix,iy,:) + w00 * line
       cube(jx,iy,:) = cube(jx,iy,:) + w10 * line
       cube(ix,jy,:) = cube(ix,jy,:) + w01 * line
       cube(jx,jy,:) = cube(jx,jy,:) + w11 * line
    else         !! pull
       line    = &
       w00*real(cube(ix,iy,:)) + &
       w10*real(cube(jx,iy,:)) + &
       w01*real(cube(ix,jy,:)) + &
       w11*real(cube(jx,jy,:))
    end if

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine weislow_interp
!----------------------------------------------------------------
  subroutine var_slow(iwb)
    integer :: iwb
    call create_modified_slowness(S%s_s%tag_local,S%s_s%tag_var,s%s_s%velocity, &
    (/s%s_s%amx%n,s%s_s%amy%n,s%s_s%az%n/), &
    (/s%s_s%amx%o,s%s_s%amy%o,s%s_s%az%o/), &
    (/s%s_s%amx%d,s%s_s%amy%d,s%s_s%az%d/), &
    var%z_samp(iwb)%dz,var%z_samp(iwb)%zloc) 
    s%s_s%tag=s%s_s%tag_var
    if(down%conv_wave) then
       call create_modified_slowness(S%s_g%tag_local,S%s_g%tag_var,s%s_g%velocity, &
       (/s%s_g%amx%n,s%s_g%amy%n,s%s_g%az%n/), &
       (/s%s_g%amx%o,s%s_g%amy%o,s%s_g%az%o/), &
       (/s%s_g%amx%d,s%s_g%amy%d,s%s_g%az%d/), &
       var%z_samp(iwb)%dz,var%z_samp(iwb)%zloc) 
       s%s_g%tag=s%s_g%tag_var
    end if
  end subroutine var_slow
  !----------------------------------------------------------------
  subroutine sellect_slow_comp(source,ith)
    logical :: source 
    integer :: ith
 
    slo%source=source
!if(run%verb)write(0,*) 'SOURCE?',source
    if(source) then
       slo%th(ith)%si=>slo%si_s
       slo%th(ith)%nrs=>slo%nrs_s
       slo%th(ith)%bound=>slo%bound_s
       slo%th(ith)%sz=>slo%sz_s

       if((down%adj) .and. (.not.down%exp_ref)) then
          down%opsign=goB2T
!if(run%verb)write(0,*) 'slow select THEN',down%exp_ref
       else
!if(run%verb)write(0,*) 'slow selectELSE',down%exp_ref
          down%opsign=goT2B
       end if
    else
       slo%th(ith)%si=>slo%si_r
       slo%th(ith)%nrs=>slo%nrs_r
       slo%th(ith)%bound=>slo%bound_r
       slo%th(ith)%sz=>slo%sz_r
       if(.not. down%adj) then
          down%opsign=goB2T
       else
          down%opsign=goT2B
       end if
    end if

  end subroutine sellect_slow_comp
  !----------------------------------------------------------------
end module wei_obj_S
