module wei_SPmig
  !! WEI migration, Shot Profile style
  use wei_util
  use wei_run
   
  use wei_obj_S
  use wei_obj_R
  use wei_obj_D
  use wei_obj_W

  use wei_fft
  use wei_kmap
  use wei_xmap

  use wei_kxmig
  use wei_image

  implicit none

  type(wedata),public :: D,D_d
  type(weimag),public :: R

  integer::d_dpresent,wpresent,w_dpresent
!  character(len=128) :: name1,name2
contains
  subroutine SPmig_init(SLin,WCin,FKin,FXin,IGin,in6,in7,in8,in9)
    integer,external,optional ::                    in6,in7,in8,in9
    integer :: st,er

#ifdef DEBUG
    character(len=128), parameter :: unit='weimig_init'
#endif
    interface !!-------------------------
       integer function    SLin() result(st)
       end function SLin
       integer function    WCin() result(st)
       end function WCin
       integer function    FKin() result(st)
       end function FKin
       integer function    FXin() result(st)
       end function FXin
       integer function    IGin() result(st)
       end function IGin
    end interface !!---------------------


    !! initialize the main WEI variables
    if(.not. weirun_init()) &
    call erexit("trouble initializing weirun_init")
    if(.not. down_init()) &
    call erexit("trouble initializing down_init")
    if(.not. init_distrib_pattern()) &
    call erexit("trouble initializing distribution pattern")


#ifdef DEBUG
    if(run%debug) call in(unit)
#endif

    !! initialize objects
    if(run%verb) call weimessage("Initialize data and source")
    if(.not. wei_data_init(D,down%D)) &
    call erexit("trouble initializing D")
    if (.not.down%passive)then
       if(.not. wei_source_init(D_d,down%D_d)) & 
       call erexit("trouble initializing D_d")
    else
       if(.not. wei_aux_data_init(D_d,down%D_d)) &
       call erexit("trouble initializing D_d")
!       call data_copy(D ,D_d)    !! src=recvr for passive
    end if

    if(run%verb) call weimessage("Initialize Slowness")
    if(.not. wei_slow_init(S,down%VS,down%VG))&
    call erexit("trouble initializing slowness")

    if (down%marine) call data_indices_marine(D,D_d)

    if(run%verb) call weimessage( "Initalizing Image ")
    if(.not. wei_image_init(R,down%R)) &
    call erexit("trouble initializing image R")

    if (0.ne.getch("W",'d',wpresent) )then
       if(down%image_freqs)call seperr("piss off, you cant have everything")
       down%wIOu=.true.
       if(.not. weibwfld_init(W  ,down%W  ))call seperr('badness at weibwfld_init(W')
       if(run%verb) write(0,*) "        *      Writing wavefield file W.    wIOu =",down%wIOu
    end if
    if (0.ne.getch("W_d",'d',w_dpresent) ) then
       down%wIOd=.true.
       if(down%image_freqs)call seperr("piss off, you cant have everything")
       if(.not. weibwfld_init(W_d,down%W_d))call seperr('badness at weibwfld_init(W_d')
       if(run%verb) write(0,*) "        *      Writing wavefield file W_d.  wIOd =",down%wIOd
    end if

!! . .  Maping files to tags
    down%mod_    = down%R     !SINGLE SHOT MODEL (REFLECTIVITY)
    down%dat_    = down%D     !DATA IS FREQUENCY SLICES
    down%dat_d_  = down%D_d   !SOURCE DATA IS FREQUENCY SLICES

    if(run%verb) call weimessage( "Initialize Fourier stuff")
    !! convert frequency to radians
      call process_radians()
    !! init FFT
    call weifft_init()  
    !! init K-map
    call weikmap_init()
    call weixmap_init()

    !! init imaging
    call weiimage_init()

    if(run%verb) call weimessage( "Allocating slowness arrays")
    !! allocate slowness arrays for FK/FX ops
    call weislow_allocate()

    if(run%verb) call weimessage( "Initializing all operators")
    !! init operators
    st=SLin()     !! FK slownesses init
    st=WCin()     !! WCop init
    st=FKin()     !! FKop init
    st=FXin()     !! FXop init
    st=IGin()     !! FXop init 
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine SPmig_init

  !----------------------------------------------------------------
  !----------------------------------------------------------------
  subroutine weimig_outnull()
  !ZERO OUTPUT IF REQUESTED
#ifdef DEBUG
    character(len=128), parameter :: unit='weimig_outnull'
    if(run%debug) call in(unit)
#endif

    if(down%adj) then !! R<-D
       call     image_assign(R,(0.,0.))
    else             !! R->D
       call      data_assign(D ,(0.,0.))
    end if

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine weimig_outnull
  !----------------------------------------------------------------
  !----------------------------------------------------------------
  integer function SPmig(adj,add,mod_,dat_,dat_d_,SLop,WCop,FKop,FXop,IGop,op6,op7,op8,op9)
  logical,            intent(in) :: adj,add
  character(len=128), intent(in) :: mod_,dat_,dat_d_
    integer,external,optional :: WCop,FKop,FXop,IGop,op6,op7,op8,op9
    integer :: izb,iwb,isx,isy
    integer :: st

#ifdef DEBUG
    character(len=128), parameter :: unit='SPmig'
#endif
    interface !!----------------------------------------------
       integer function SLop() result(st)
       end function SLop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%debug) call in(unit)
#endif

    !! reset tag variables
    call down_reset2(adj,add,mod_,dat_,dat_d_)
!    if(run%verb) call process_follow() !! where am I?
    
    !! accomodate tag name changes
#ifdef SEP_MPI
    R%tag_0     = down%mod_
    D%tag_0     = down%dat_
    D_d%tag_0   = down%dat_d_
#else
    R%tag     = trim(down%mod_)
    D%tag     = trim(down%dat_)
    D_d%tag   = trim(down%dat_d_)
#endif
    !! setup the adjoint sign
    if(down%adj) then
       down%opsign = goT2B
    else
       down%opsign = goB2T
    end if

    !! zero the output if add is off
!    if(.not. down%add) call weimig_outnull() !uneseccary, zeroed during initialize

    down%shotnum=0

    shoty: do isy=1,down%asy%n
    shotx: do isx=1,down%asx%n
       if (run%verb)then
          call weiline()
          write(0,'(a,i5,a,i5)')"         *  Shotnum",&
          down%shotnum+1,' of',down%asy%n*down%asx%n
       end if
       if(down%marine) then
          call data_position(D,isx,isy) 
          call wei_xmap()
       else
          call data_indices_land(D_d,isx,isy)
       end if
       call data_position_report()
       if( .not.down%passive)&
       call source_report(D_d)

       SPmig=SPmigof(isx,isy,SLop,WCop,FKop,FXop,IGop)

#ifdef SEP_MPI 
       if (down%wIOd .and. run%mpi) then
          if(run%verb) call weimessage("COMBINE DOWN WAVEFIELD")
          call combine_wavefield(W_d)
       end if
       if (down%wIOu .and. run%mpi) then
          if(run%verb) call weimessage("COMBINE UP WAVEFIELD")
          call combine_wavefield(W)
       end if
#endif
       if(run%verb)write(0,'(a,i4)') '         *       Finished migrating shot',down%shotnum +1

       down%shotnum=down%shotnum+1
       st=sseek(D_d%tag,0,0)               !! rewind to beginning of source
    end do shotx
    end do shoty

#ifdef SEP_MPI 
    if(down%wIOu)call aux_unlink(trim(W%tag))
    if(down%wIOd)call aux_unlink(trim(W_d%tag))
    if(run%verb) call weimessage("COMBINE SHOT IMAGES")
    call combine_image(R)
    call aux_unlink(D_d%tag)
    call aux_unlink(D%tag)
    call aux_unlink(S%s_s%tag)
    if(down%conv_wave) call aux_unlink(S%s_g%tag)
    if(run%verb) call weimessage("OK")
#endif

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    SPmig=OK
    
  end function SPmig

!--------------------------------------------------------------------
  integer function SPmigof(isx,isy,SLop,WCop,FKop,FXop,IGop)
    integer,external,optional :: WCop,FKop,FXop,IGop
    integer :: izb,iwb,position
    integer :: st,ierr,isx,isy
#ifdef DEBUG
    character(len=128), parameter :: unit='weimigof'
#endif
    interface !!----------------------------------------------
       integer function SLop() result(st)
       end function SLop
    end interface !!------------------------------------------
    var%dzb=down%az%d

    if(down%adj) then !! MIGRATION
#ifdef DEBUG
       if(run%debug) call in('migration')
#endif 
!       call to_history("fblock",0,R%tag)        
       wbA: do iwb=1,down%nwb
          if(run%verb)write(0,'(a,i2,a,i2)') '         *       Frequency block ',iwb,' of',down%nwb
          position=down%shotnum*down%nwb + (iwb-1)

          if(.not. data_read(D,position)) &
          call seperr("trouble reading D data") !! read data at block top
          if (.not.down%passive ) then !!D_d already pointed to D
             if(.not. data_read(D_d,iwb-1)) & !!read the source file
             call seperr("trouble reading D_d data")
          else
             call data_copy(D,D_d)
          end if

          zbA: do izb=1,down%nzb       
             call setup_var_block(iwb,izb)
             R%bin=0.;
             call slowness_read(S,izb) !! read a slowness block
             st = SLop() !! setup sS, sR, s(z) for working space
             down%bw%o = down%aw%o + (down%iw_first(iwb)-1)*down%bw%d
             if(down%wIOu .or. down%wIOd) then           !! DC one block
                call wmigWW(izb,iwb,WCop,FKop,FXop,IGop) !! write W
             else
                call wmigRD(izb,iwb,down%nzs,WCop,FKop,FXop,IGop) !! trash W
             end if
             if (.not. down%image_freqs) then
                if (.not. image_insert(R))&
                call seperr("Single shot image => global image didnt go well")
!                call image_write(R,izb,(iwb/=1 .or. down%shotnum/=0 .or. down%add .or. run%share_image))
                call image_write(R,izb,.false.)
             end if
          end do zbA
       end do wbA

       !----------------------------------------------------------
#ifdef DEBUG
       if(run%debug) call out('migration')
#endif
    else         !! MODELING
#ifdef DEBUG
       if(run%debug) call in('modeling')
#endif
       !----------------------------------------------------------

#ifdef DEBUG
       if(run%debug) call out('modeling')
#endif

       !! no modeling operator yet

    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    SPmigof=OK
  end function SPmigof 
  !---------------------------------------------------------------
  !! don't write the wavefield
  subroutine wmigRD(izb,iwb,nzs,WCop,FKop,FXop,IGop)
    integer, intent(in)       :: izb,iwb
    integer,external,optional :: WCop,FKop,FXop,IGop 
    integer :: izs,iws,ith,nzs
#ifdef DEBUG
    character(len=128), parameter :: unit='wmigRD'
    if(run%debug) call in(unit)
#endif
    if(down%adj) then !! MIGRATION
       if (down%image_freqs) ith=sseek(R%tag,0,0)
       ith=1
       zsA: do izs=1,nzs          !LOOP OVER z STEPS
          if (izs .ge. down%firstz)& !! limit file size when min. target depth known
          call SPimg(izs,R,D%imgbin,D_d%imgbin,IGop)

          if (mod(izs,down%az%n/10)==0) call depth_report(izs,izb,iwb)
          if (izs==down%nzs) return !! no need to DC further than the model space

          if(down%antion) D%imgbin=(0.,0.) !zero imaging wavefield
          if(down%antion .and. (.not.down%passive)) D_d%imgbin=(0.,0.)

          wsA: do iws=1,down%nws             !LOOP OVER w STEPS
!          if(down%antion) D%imgbin(1:size(D%imgbin,1),1:size(D%imgbin,2),iws)=(0.,0.) !zero imaging wavefield
!          if(down%antion .and. (.not.down%passive)) D_d%imgbin(1:size(D%imgbin,1),1:size(D%imgbin,2),iws)=(0.,0.)
          call weikxmig_s(D%bin,D%imgbin,D_d%bin,D_d%imgbin,iws,izs,ith,WCop,FKop,FXop)
          end do wsA
       end do zsA
    else         !! MODELLING
       call erexit('Shot profile modeling not yet implemented.')
!       wsF: do iws=1,down%nws             !LOOP OVER w STEPS
!          ith=1
!          zsF: do izs=down%nzs,1,-1       !LOOP OVER z STEPS
             !!need to write weiimgSPadj
             !!until then, do nothing
!             call weikxmig(bW%bin,bW_d%bin,iws,izs,ith,WCop,FKop,FXop)
!             call weiimgSPadj(izs,iws,ith, R%bin,bW%bin,IGop)
!          end do zsF
!       end do wsF
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine wmigRD
  !----------------------------------------------------------------
  !! write the wavefield
  subroutine wmigWW(izb,iwb,WCop,FKop,FXop,IGop)
    integer, intent(in)       :: izb,iwb
    integer,external,optional :: WCop,FKop,FXop,IGop
    integer :: izs,iws,ith
#ifdef DEBUG
    character(len=128), parameter :: unit='wmigWW'
    if(run%debug) call in(unit)
#endif

    if(down%adj) then !! MIGRATION
       ith=1
       if (down%antion) D%imgbin=D%bin; D_d%imgbin=D_d%bin
       zsA: do izs=1,down%nzs
          if (down%wIOu)then
             if (.not. wavefield_write(W  ,D%bin  ,iwb,izb,izs)) &
             call seperr('wavefield_write(W ==badness')
          end if
          if (down%wIOd)then
             if (.not. wavefield_write(W_d,D_d%bin,iwb,izb,izs)) &
             call seperr('wavefield_write(W_d ==badness')
          end if

          call SPimg(izs,R,D%imgbin,D_d%imgbin,IGop)

          if (mod(izs,down%az%n/10)==0) call depth_report(izs,izb,iwb)
          if (izs==down%nzs) return
          wsA: do iws=1,down%nws
             if(down%antion) D%imgbin(1:size(D%imgbin,1),1:size(D%imgbin,2),iws)=(0.,0.) !zero imaging wavefield
             if(down%antion .and. (.not.down%passive)) D_d%imgbin(1:size(D%imgbin,1),1:size(D%imgbin,2),iws)=(0.,0.)
             call weikxmig_s(D%bin,D%imgbin,D_d%bin,D_d%imgbin,iws,izs,ith,WCop,FKop,FXop)
          end do wsA
        end do zsA
    else         !! MODELLING
       call seperr("SP modelling not implented")
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine wmigWW
  !----------------------------------------------------------------

subroutine depth_report(izs,izb,iwb)
integer :: izs,izb,iwb
integer :: pcent,current,total

pcent=(1. + 1.0*(izs+down%nzs*(izb-1))/down%az%n*100/(real(down%nwb)-(iwb-1)))
current=izs+down%nzs*(izb-1)
total=down%nzs*down%nzb
if(run%verb) then
   write(0,'(a,i5,a,i5,a,i2,a)')'        *  Finished depth number',&
      current,' of',total,' (',pcent,'%)'
end if
end subroutine depth_report
!----------------------------------------------------------------
end module wei_SPmig
















































