module wei_obj_W
  use wei_util
  use wei_down
  use wei_run
  use wei_run
  use wei_external
  use wei_type

!  use wei_var
!  use wei_type
!  use wei_obj_D
  !---------------
  implicit none

  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!
  type(wewfld), public :: W,W_d
  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!

contains
  !----------------------------------------------------------------
  logical function weibwfld_init(X,tag)
    type(wewfld),       intent(inout) :: X
    character(len=128), intent(in)    :: tag

    weibwfld_init=F
    X%tag = tag
    select case(run%operator(1:5))
    case('.R_.D')
       if(.not. wavefield_param_init(X,tag))then
          write(0,*) "trouble initializing wavefield by parameters"
          return
       end if

    case('dS_dR')
       if(isfile(X%tag)) then
          if(.not.  wavefield_history_init(X)) then
             write(0,*) "trouble initializing wavefield by history"
             return
          end if
       else
          call seperr('W /= file')
       end if
       
    case default 
       call erexit("this is a fuckup- you should never get here (weiobjW.F90)")
    end select
    weibwfld_init=T
  end function weibwfld_init
  !----------------------------------------------------------------
  logical function wavefield_history_init(X)
    type(wewfld), intent(inout) :: X
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_history_init'
    if(run%debug) call in(unit)
#endif
    wavefield_history_init=F
    call auxinout(X%tag)
    X%from = "history"

    call hinaxis(X%tag,X%amx,1); X%amx%l='mx'
       down%amx = X%amx   
    call hinaxis(X%tag,X%amy,2); X%amy%l='my'
       down%amy = X%amy
    call hinaxis(X%tag,X%aw ,3); X%aw%l='w_'
       down%aw = X%aw
    call hinaxis(X%tag,X%az ,4); X%az%l ='z_'
       down%az = X%az
    call hinaxis(X%tag,X%asx,5); X%asx%l='sx'
       down%asx = X%asx
    call hinaxis(X%tag,X%asy,6); X%asy%l ='sy'
       down%asy = X%asy

    if(.not. wavefield_setup(X,.true.)) call seperr('wavefield_setup = notgood')
    wavefield_history_init=T
#ifdef DEBUG
    if(run%debug) call wavefield_report(X)
    if(run%debug) call out(unit)
#endif
  end function wavefield_history_init
  !----------------------------------------------------------------
  logical function create_wavefield_file(X,tag)
    type(wewfld), intent(in) :: X
    character(len=*)         :: tag
       create_wavefield_file=F 
       call auxscr(tag)
       call to_history("esize",8,tag)
       call houaxes(tag,X%amx,X%amy,X%aw,X%az,X%asx,X%asy)
       create_wavefield_file=T 
     end function create_wavefield_file
!----------------------------------------------------------------
   logical function wavefield_param_init(X,tag)
    type(wewfld), intent(inout) :: X
    character(len=128), intent(in)    :: tag
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_param_init'
    if(run%debug) call in(unit)
#endif
    wavefield_param_init=F

    X%from = "parameter"
    X%amx = down%amx;    X%amy = down%amy;
    X%asx = down%asx;    X%asy = down%asy;
    X%aw  = down%aw;     X%az  = down%az; 

    if(run%m_ith==0) then 
      if(.not. create_wavefield_file(X,X%tag)) then
        write(0,*) "trouble creating image"
        return
       end if
    end if

    if(.not.wavefield_setup(X,.true.)) then
      write(0,*) "trouble at wavefield_setup",X%tag
      return
    end if

   wavefield_param_init=T
#ifdef DEBUG
    if(run%debug) call wavefield_report(X)
    if(run%debug) call out(unit)
#endif
  end function wavefield_param_init
  !----------------------------------------------------------------
  logical function wavefield_setup(X,zero)
    type(wewfld), intent(inout) :: X
    logical :: zero
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_setup'
    if(run%debug) call in(unit)
#endif
    wavefield_setup=F

    X%bites = 8* product( (/X%amx%n,X%amy%n,down%nws/) )

    !! DEPTH blocking
    if( (1.0*X%az%n)/down%nzs /= X%az%n/down%nzs ) then
       call wavefield_report(X)
       call seperr('W.az.n is not a multiple of W.nzs')
    end if

    !! FREQUENCY blocking
    if( (1.0*X%aw_local%n)/down%nws /= X%aw_local%n/down%nws ) then
       call wavefield_report(X)
       call seperr('W.aw.n is not a multiple of W.nws')
    end if

  if(run%mpi) then
     if(.not.  mpi_wavefield_setup(X)) then
       write(0,*) "trouble at mpi_wavefield_setup(",X%tag
       return
     end if
   else
     X%aw_local=X%aw
   end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    wavefield_setup=T
  end function wavefield_setup
  !----------------------------------------------------------------
  !
  !   Setup wavefield structure for local files if
  !     we are running MPI
  !
  !

  logical function mpi_wavefield_setup(X)
    type(wewfld), intent(inout) :: X
    integer ::n,i
    mpi_wavefield_setup=F

    X%tag_0=X%tag !! move command line R+ to output tag
    !! make tags for each processor because pgf90 sucks ass
    write(X%tag,*) trim(X%tag_0),trim(run%tag_num)
    i=1
    do while(i<128 .and. X%tag(i:i)==" ")
      i=i+1
    end do
    X%tag=X%tag(i:len(trim(X%tag)))

    X%aw_local%n=down%nw_local*down%nws
    n=X%aw%n
    X%aw%n=down%nw_local*down%nws
    if(.not. create_wavefield_file(X,X%tag)) then
      write(0,*) "MPI trouble at create_wavefield_file(",X%tag
      return
    end if
    X%aw%n=n
!       if(.not. file_val_assign(X%tag,cmplx(0.,0.))) then
!          write(0,*) "trouble setting ",trim(X%tag_0)
!          return
!       end if

    mpi_wavefield_setup=T
  end function mpi_wavefield_setup
  !----------------------------------------------------------------
  subroutine wavefield_assign(X,v)
    type(wewfld)        :: X
    complex, intent(in) :: v
    logical ::a
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_assign'
    if(run%debug) call in(unit)
#endif
    a=file_val_assign(X%tag,v)

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine wavefield_assign
  !----------------------------------------------------------------  
  ! put 'em all back together after mpi runs
  subroutine combine_wavefield(X)
    type(wewfld), intent(inout) :: X
    integer :: ierr,i,iz,k
    integer, external :: sseek
#ifdef SEP_MPI

!    k=down%shotnum*X%bites*X%az%n
    i=0;if(down%add) i=1
    if( down%adj .and. run%mpi) then
       if((run%m_ith==0) .and. (down%shotnum==0)) &
!         ierr= sseek(trim(x%tag_0),k,0)
         ierr= sseek(trim(x%tag_0),0,0)
         ierr= sseek(trim(X%tag  ),0,0)
       do iz=1,down%az%n
          if(0/=mpi_sep_tag_combine(run%m_ith,X%tag,0,X%tag_0, &
           X%bites/4,down%nwp,down%iw_thread,i)) then
             write(0,*) "trouble combining wavefield files",iz
             return
          end if
       end do
    end if
#endif
  end subroutine combine_wavefield

  !----------------------------------------------------------------  
  subroutine wavefield_report(X)
    type(wewfld), intent(in) :: X

    write(0,*)'        ************************************************'
    write(0,*)'        *       ',trim(X%tag),' from ',trim(X%from)
    write(0,'(a,i3,f10.2,f10.3)       ') "         *   Offx: ",X%amx%n, X%amx%o, X%amx%d
    write(0,'(a,i3,f10.2,f10.3)       ') "         *   Offy: ",X%amy%n, X%amy%o, X%amy%d
    write(0,'(a,i3,f10.2,f10.3)       ') "         *   Shtx: ",X%asx%n, X%asx%o, X%asx%d
    write(0,'(a,i3,f10.2,f10.3)       ') "         *   Shty: ",X%asy%n, X%asy%o, X%asy%d
    write(0,'(a,i3,f10.2,f10.3,a,i3,a)') "         *     W: ",X%aw%n,  X%aw%o,  X%aw%d," ->",down%nwb," blocks"
    write(0,'(a,i3,f10.2,f10.3,a,i3,a)') "         *     Z: ",X%az%n,  X%az%o,  X%az%d," ->",down%nzb," blocks"
    write(0,*)'        ************************************************'
  end subroutine wavefield_report
  !--------------------------------------------------------------
  logical function wavefield_read(X,data,iwb,izb,izs)
    type(wewfld), intent(inout) :: X
    complex,dimension(:,:,:), intent(in)   :: data
    integer,intent(in) :: iwb,izb,izs
    integer :: stat,k
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_read'
    if(run%depar) call in(unit)
#endif
    wavefield_read=F
    k=(iwb-1)                   &
    + (izb-1)*down%nwb*down%nzs &
    + (izs-1)*down%nwb

    stat=sseek_block(X%tag,k,X%bites,0)
    if(X%bites /=sreed(X%tag, data, X%bites))then
       write(0,*) "trouble reading wavefield block",k,'machine',run%m_ith
    else
       wavefield_read=T
    end if
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end function wavefield_read
  !----------------------------------------------------------------
  logical function wavefield_write(X,data,iwb,izb,izs)
    type(wewfld), intent(inout) :: X
    complex,dimension(:,:,:), intent(in)   :: data
    integer,intent(in):: iwb,izb,izs
    integer :: stat,k,n
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_write'
    if(run%depar) call in(unit)
#endif
    wavefield_write=F
    if(run%mpi) then
       n=X%aw%n
       X%aw%n=down%nw_local*down%nws
    end if

    k= (iwb-1)                   &
     + (izb-1)*down%nwb*down%nzs &
     + (izs-1)*down%nwb 

    if(run%mpi) then
       X%aw%n=n
    end if

    stat=sseek_block(X%tag,k,X%bites,0)
    if(X%bites /= srite(X%tag, data, X%bites))then
       write(0,*) "trouble writing wavefield block",k
       return
    end if
    call to_history("ifreq",k,X%tag)

       wavefield_write=T
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end function wavefield_write
  !----------------------------------------------------------------
end module wei_obj_W









