module wei_image
  !! WEI imaging condition

  use wei_util
  use wei_down
  use wei_down_type
  use wei_type
  use wei_run
  !---------------
  implicit none

  complex, dimension(:,:,:,:), pointer, private :: imwf
contains
  !----------------------------------------------------------------  
  subroutine weiimage_init()
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weiimage_init'
    if(run%debug) call in(unit)
#endif

    !! imaging wavefield
    allocate(imwf(down%amx%n,down%amy%n,down%ahx%n,down%ahy%n),stat=er)
    if(er/=0) call seperr('cant allocate imwf for shot profile')
    imwf=cmplx(0.,0.)
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine weiimage_init
!------------------------------------------------------------------------------------------
  subroutine SPimg(izs,R,wfld_u,wfld_d,IGop)
    integer, intent(in) :: izs
    type(weimag),intent(inout) :: R
    complex, dimension(:,:,:), pointer :: wfld_u,wfld_d
    integer :: st
#ifdef DEBUG
    character(len=128), parameter :: unit='SPimg'
#endif
    interface !!-------------------------
       integer function IGop(image,wfld_u,wfld_d,R) result(st)
         use wei_type
         complex, dimension(:,:,:), pointer :: wfld_u,wfld_d,image
         type(weimag), intent(inout) ::R
       end function IGop
    end interface !!---------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif

  !! image all offsets for all frequencies
  st = IGop(imwf,wfld_u,wfld_d,R)
  !! sum up frequency blocks from temp model into actual model
if (.not.down%image_freqs) then
  if (down%passive) then
     R%bin(:,:,:,:,izs)=R%bin(:,:,:,:,izs) + imwf(:,:,:,:) !/down%aw%n
  else
     R%bin(:,:,:,:,izs)= imwf(:,:,:,:) /down%aw%n !uncommented division
  end if
end if
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine SPimg
end module wei_image





