module wei_kmap
  !! K-domain mapping

  use wei_run
  use wei_down
  use wei_external
  use wei_type
  !---------------
  implicit none
  integer, private, save :: nhx_use,nhy_use

  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!
  type(wekmap), public :: kmap
  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!

contains
  !----------------------------------------------------------------
  subroutine weikmap_init()
    integer :: imx,imy
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weikmap_init'
    if(run%debug) call in(unit)
#endif

    allocate(kmap%kx(down%amx%n),stat=er)
    if(er/=0) call seperr('cant allocate kx')
    call klinemap(kmap%kx,down%amx%n,down%amx%d,0)

    allocate(kmap%ky(down%amy%n),stat=er)
    if(er/=0) call seperr('cant allocate ky')
    call klinemap(kmap%ky,down%amy%n,down%amy%d,0)

    if(run%verb)call weikmap_report()
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine weikmap_init
!----------------------------------------------------------------
  subroutine weikmap_report()
  write(0,*)"        *  Wavenumber report"
  write(0,*)"        *    kx",size(kmap%kx),kmap%kx(1),kmap%kx(2)-kmap%kx(1)
  if (size(kmap%ky) .gt. 1) then
  write(0,*)"        *    ky",size(kmap%ky),kmap%ky(1),kmap%ky(2)-kmap%ky(1)
  end if
  end subroutine weikmap_report
  !----------------------------------------------------------------
  subroutine klinemap(k,n,d,shift)
    real, dimension(:), pointer :: k
    integer, intent(in) :: n,shift
    real,    intent(in) :: d
    integer             :: i,m

    k=0.
    if(n/=1) then
       m = n/2
       do i=1,n
          if(i <= m) then
             k(i+shift) = i -1
          else
             k(i+shift) = i - n -1
          end if
       end do
       !       do i=1,m
       !          k(i+shift  ) =    i-1
       !          k(i+m+shift) = -m+i-1
       !       end do
    else
       k=0. !! make sure k=0 if n=1
    end if
    k =  k / (n*d)  !! 1/length
    k = 2 * pi * k !! radians

  end subroutine klinemap
  !----------------------------------------------------------------
end module wei_kmap
