module wei_kxmig
  !! fk-fx migration
  !! receive wfld at (izs,iws,ith)
  !! return  wfld downward/upward continued
  !! needs Fop, FXop

  use wei_util
  use wei_run
  use wei_external
  use wei_down
  use wei_obj_S
  use wei_fft
  use wei_kmap
  !---------------
  implicit none


  !! temporary interpolation arrays
  complex, dimension(:,:,:), pointer, private :: tmp1,tmp2
  complex, dimension(:,:), pointer, private :: tmps
  character(len=128) :: twave

contains
  !----------------------------------------------------------------
  !----------------------------------------------------------------
  subroutine weikxmig_s(wfld_r,wfld_r_img,wfld_s,wfld_s_img,iws,izs,ith,KXMIGop,FKop,FXop)
    complex, dimension(:,:,:), pointer         :: wfld_s,wfld_r
    complex, dimension(:,:,:), pointer         :: wfld_s_img,wfld_r_img
    integer,external,optional ::    FKop,FXop
    integer :: izs,iws,ith
    integer :: st
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmig'
#endif
    interface !!-------------------------
       integer function KXMIGop(wfld_r,wfld_r_img,wfld_s,wfld_s_img,iws,izs,ith,FKop,FXop) result(st)
         complex, dimension(:,:,:), pointer ::wfld_s,wfld_r
         complex, dimension(:,:,:), pointer ::wfld_s_img,wfld_r_img
         integer, intent(in)       :: iws,izs,ith
         integer,external,optional :: FKop,FXop
       end function KXMIGop
    end interface !!---------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif

    st = KXMIGop(wfld_r,wfld_r_img,wfld_s,wfld_s_img,iws,izs,ith,FKop,FXop)

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine weikxmig_s

 
  !----------------------------------------------------------------

  !----------------------------------------------------------------
  !! N reference slownesses
  integer function weikxmigN_init() result(st)
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmigN_init'
    if(run%debug) call in(unit)
#endif

      !INITIALIZE TEMPORARY SPACE FOR WAVEFIELDS 
      call init_wave_temps()
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weikxmigN_init

  !----------------------------------------------------------------
  !! N reference slownesses
  integer function weikxmigN_shot(wfld_rec,wfld_rec_img,wfld_src,wfld_src_img,iws,izs,ith,FKop,FXop) result(st)
    complex, dimension(:,:,:), pointer :: wfld_src,wfld_rec
    complex, dimension(:,:,:), pointer :: wfld_src_img,wfld_rec_img
    integer, intent(in)       :: iws,izs,ith
    integer                   :: ifk,isource,ierr
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmigN'
#endif
    interface !!----------------------------------------------
       integer function FKop(iws,izs,ifk,ith,wfld,wfld_img) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:), pointer :: wfld,wfld_img
       end function FKop
       integer function FXop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:), pointer :: wfld
       end function FXop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif

      !!D.C. SOURCE WAVEFIELD
!      call sellect_slow_comp(.true.,ith) !!selects slowness from source field
      call sellect_slow_comp(.true.,1) !!selects slowness from source field
      st=weikxmigN_single(wfld_src,wfld_src_img,iws,izs,ith,Fkop,Fxop)
      !!D.C. RECEIVER WAVEFIELD
!      call sellect_slow_comp(.false.,ith)!!selects slowness from recvr field
      call sellect_slow_comp(.false.,1)!!selects slowness from recvr field
      st=weikxmigN_single(wfld_rec,wfld_rec_img,iws,izs,ith,Fkop,Fxop)

    end function weikxmigN_shot
!----------------------------------------------------------------
  !! N reference slownesses
  integer function weikxmigN_single(wfld,wfld_img,iws,izs,ith,FKop,FXop) result(st)
    complex, dimension(:,:,:), pointer     :: wfld,wfld_img
    integer, intent(in)                    :: iws,izs,ith
    integer                                :: ifk,isource,ierr
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmigN'
#endif
    interface !!----------------------------------------------
       integer function FKop(iws,izs,ifk,ith,wfld,wfld_img) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:), pointer :: wfld,wfld_img
        end function FKop
       integer function FXop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:), pointer :: wfld
       end function FXop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif
    if(down%adj) then !! MIGRATION {
       !----------------------------------------------------------------       
       call store_wavefield(wfld,ith,iws) !STORE THE CURRENT WAVEFIELD
       do ifk=1,slo%th(ith)%nrs(izs) !{
         call grab_part_wvfld(wfld,ifk,ith,iws,izs) !GRAB A PORTION OF THE WFLD
          st   = FXop(iws,izs,ifk,ith,wfld) !! f-x operator
          call weiift(iws,izs,    ith,wfld)
          st   = FKop(iws,izs,ifk,ith,wfld,wfld_img) !! f-k operator
          call weifft(iws,izs,    ith,wfld)
          call add_wavefield(wfld,ith,iws) !SUM OUR EFFORTS TOGETHER
       end do !}
        call grab_wavefield(wfld,iws,ith,.true.) !GRAB THE ENTIRE WAVEFIELD
        if(down%antion)then
           call weifft(iws,izs,    ith,wfld_img)
        end if
       !----------------------------------------------------------------
  !}
    else             !! MODELLING {
       !----------------------------------------------------------------     
    call sellect_slow_comp(isource==1,ith)
        call store_wavefield(wfld,iws,ith)

       do ifk=1,slo%th(ith)%nrs(izs) !{
          call grab_wavefield(wfld,iws,ith,.false.)
          call weiift(iws,izs,    ith,wfld)
          st   = FKop(iws,izs,ifk,ith,wfld,wfld_img) !! f-k op
          call weifft(iws,izs,    ith,wfld)
          st   = FXop(iws,izs,ifk,ith,wfld) !! f-x op
          call store_part_wavefield(wfld,ifk,ith,iws,izs) 
       end do !}
        call grab_wavefield(wfld,iws,ith,.true.)
       !----------------------------------------------------------------
  end if!}
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weikxmigN_single
  !----------------------------------------------------------------  
!! multiple fk ops
!!
!!       D(z)
!!       |
!!       v
!!       T1
!!
!!    D1 D2 D3 ...
!!
!!    S1 S2 S3 ...
!!     \ | /
!!       T2
!!       |
!!       v
!!       D(z+dz)
!---------------------------------------------
!  Initialize storage for wavefields
subroutine init_wave_temps() 
 integer ::er

   !! allocate tmp arrays for interpolation of multiple fk ops
   allocate(tmp1(down%amx%n,down%amy%n,run%o_nth),stat=er)
   if(er/=0) call seperr('cant allocate tmp1')
   allocate(tmp2(down%amx%n,down%amy%n,run%o_nth),stat=er)
   if(er/=0) call seperr('cant allocate tmp2')

 end subroutine init_wave_temps

!-------------------------------------------------
! Store the wavefield
subroutine store_w(wfld)
 complex :: wfld(:,:,:)
 integer :: ierr
  ierr= sseek(twave,0,0)
  ierr= srite(twave,wfld(:,:,1),8*size(wfld(:,:,1)))
end subroutine store_w
!----------------------------------------------------------------
subroutine store_wavefield(wfld,ith,iws)
 complex :: wfld(:,:,:)
 integer :: ierr,iws,ith,i1,i2

 tmp2(:,:,ith) = wfld(:,:,iws)
 tmp1=0.

end subroutine store_wavefield
!----------------------------------------------------------------
!subroutine zero_imgbin(wfld,iws)
subroutine zero_imgbin(wfld)
 complex :: wfld(:,:,:)
 integer :: iws

! wfld(:,:,iws) = (0.,0.)
 wfld(:,:,:) = (0.,0.)

end subroutine zero_imgbin
!------------------------------------------------------
subroutine grab_part_wvfld(wfld,ifk,ith,iws,izs)
  integer :: ith,iws,ifk,ierr,izs
  complex :: wfld(:,:,:)

     where(slo%th(ith)%si(:,:,izs) == ifk)
        wfld(:,:,iws) = tmp2(:,:,ith)
     elsewhere
        wfld(:,:,iws) = 0.
     end where

end subroutine grab_part_wvfld
!------------------------------------------------------
subroutine add_wavefield(wfld,ith,iws)
complex :: wfld(:,:,:)
integer :: ith,iws,i1,i2,ierr

  tmp1(:,:,1) = tmp1(:,:,ith) +  wfld(:,:,iws)

end subroutine add_wavefield

!----------------------------------------------------------------
subroutine grab_w(wfld)
complex :: wfld(:,:,:)
logical :: first
integer :: ith,iws,i1,i2,ierr
  ierr= sseek(twave,0,0)
  ierr= sreed(twave,wfld(:,:,1),8*size(wfld(:,:,1)))
end subroutine grab_w
!----------------------------------------------------------------
subroutine grab_wavefield(wfld,iws,ith,first)
  complex :: wfld(:,:,:)
  logical :: first
  integer :: ith,iws,i1,i2,ierr

  wfld(:,:,iws) = tmp1(:,:,ith)
end subroutine grab_wavefield

!----------------------------------------------------------------
subroutine store_part_wavefield(wfld,ifk,ith,iws,izs)
  complex wfld(:,:,:)
  integer :: ifk,ith,iws,ierr,izs,i1,i2

  where(slo%si(:,:,izs) == ifk)
     tmp1(:,:,ith) = tmp1(:,:,ith) + wfld(:,:,iws)
  end where
end subroutine store_part_wavefield
!----------------------------------------------------------------
end module wei_kxmig




