module wei_sll
  !! WEI slowness operator
  !! for FK-FX downward continuation
  !! includes salt logic

  use wei_util
  use wei_obj_S
  use lloyd_mod
  !---------------
  implicit none

  real, dimension(:,:), pointer, private :: sxy
  real, dimension(:), pointer, private :: bound,ref_slow,s_slice

contains
  !----------------------------------------------------------------
  !! N reference slownesses
  integer function weisll_init() result(st)
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weislsN_init'
    if(run%debug) call in(unit)
#endif
    allocate(bound(down%nref),ref_slow(down%nref))
    allocate(s_slice(size(S%s_s%bin,1)*size(S%s_s%bin,2)))
    allocate(sxy(down%amx%n,down%amy%n),stat=er)
    if(er/=0) call seperr('cant allocate sxy')
    call init_lloyd()
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weisll_init


  !----------------------------------------------------------------  
  !! N reference slownesses
  integer function weisll() result(st)
    integer :: izs,ifk,i2,i1
    real    :: ds, so, sz
    real    :: sl,su
    real    :: smin,smax
    integer :: imx, imy
    real    :: ssalt_true
    real    :: ssalt_mean,ssalt_sum
    real    :: slocal
    integer :: ins,ndo,ierr

#ifdef DEBUG
    character(len=128), parameter :: unit='weislsN'
    if(run%debug) call in(unit)
#endif

    !! spread the slowness in its PBA    
    call weislow_push(S)
    call sellect_slow_comp(.true.,1)
    do izs=1,down%nzs
       slo%th(1)%nrs(izs) = down%nref
       sxy(:,:) =real(S%pba(:,:,izs))
       s_slice=reshape(S%s_s%bin(:,:,izs),(/size(S%s_s%bin(:,:,izs))/))
       call lloyd_setup(izs) 
    end do

    if (down%conv_wave) then
       call sellect_slow_comp(.false.,1)
       do izs=1,down%nzs
          slo%th(1)%nrs(izs) = down%nref
          sxy(:,:) =aimag(S%pba(:,:,izs))
          s_slice=reshape(S%s_g%bin(:,:,izs),(/size(S%s_g%bin(:,:,izs))/))
          call lloyd_setup(izs) 
       end do
       else 
          slo%nrs_r=slo%nrs_s
          slo%sz_r =slo%sz_s
          slo%si_r =slo%si_s
          slo%bound_r=slo%bound_s
       end if

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weisll
  !----------------------------------------------------------------
  subroutine lloyd_setup(izs)
    integer :: i1,i2,izs

    slo%th(1)%nrs(izs)=lloyd_go(s_slice,ref_slow,down%nref,bound,run%m_ith==0)
    slo%th(1)%sz(izs,1:size(ref_slow)) =ref_slow
    slo%th(1)%bound(izs,1:size(bound))= bound

    do  i2=1,size(sxy,2)
       do  i1=1,size(sxy,1)
          slo%th(1)%si(i1,i2,izs)=1
          do while(sxy(i1,i2) > bound(slo%th(1)%si(i1,i2,izs)))
             slo%th(1)%si(i1,i2,izs)=slo%th(1)%si(i1,i2,izs)+1
          end do
       end do
    end do
  end subroutine lloyd_setup
!----------------------------------------------------------------
end module wei_sll




